from pathlib import Path
from typing import Iterator, List, Sequence, Tuple, Union, cast

from npe2.manifest import PluginManifest, contributions
from npe2.manifest.utils import merge_contributions, merge_manifests

from ._setuputils import get_package_dir_info
from ._visitors import find_npe2_module_contributions


def find_packages(where: Union[str, Path] = ".") -> List[Path]:
    """Return all folders that have an __init__.py file"""
    return [p.parent for p in Path(where).resolve().rglob("**/__init__.py")]


def get_package_name(where: Union[str, Path] = ".") -> str:
    return get_package_dir_info(where).package_name


def compile(
    src_dir: Union[str, Path],
    dest: Union[str, Path, None] = None,
    packages: Sequence[str] = (),
    plugin_name: str = "",
    template: Union[str, Path, None] = None,
) -> PluginManifest:
    """Compile plugin manifest from `src_dir`, where is a top-level repo.

    This will discover all the contribution points in the repo and output a manifest
    object

    Parameters
    ----------
    src_dir : Union[str, Path]
        Repo root. Should contain a pyproject or setup.cfg file.
    dest : Union[str, Path, None]
        If provided, path where output manifest should be written.
    packages : Sequence[str]
        List of packages to include in the manifest.  By default, all packages
        (subfolders that have an `__init__.py`) will be included.
    plugin_name : str
        Name of the plugin. If not provided, the name will be derived from the
        package structure (this assumes a setuptools package.)
    template : Union[str, Path, None]
        If provided, path to a template manifest file to use. This file can contain
        "non-command" contributions, like `display_name`, or `themes`, etc...
        In the case of conflicts (discovered, decoratated contributions with the same
        id as something in the template), discovered contributions will take
        precedence.

    Returns
    -------
    PluginManifest
        Manifest including all discovered contribution points, combined with any
        existing contributions explicitly stated in the manifest.
    """
    src_path = Path(src_dir)
    assert src_path.exists(), f"src_dir {src_dir} does not exist"

    if dest is not None:
        pdest = Path(dest)
        suffix = pdest.suffix.lstrip(".")
        if suffix not in {"json", "yaml", "toml"}:
            raise ValueError(
                f"dest {dest!r} must have an extension of .json, .yaml, or .toml"
            )

    if template is not None:
        template_mf = PluginManifest.from_file(template)

    _packages = find_packages(src_path)
    if packages:
        _packages = [p for p in _packages if p.name in packages]

    if not plugin_name:
        plugin_name = get_package_name(src_path)

    contribs: List[contributions.ContributionPoints] = []
    for pkg_path in _packages:
        top_mod = pkg_path.name
        # TODO: add more tests with more complicated package structures
        # make sure we're not double detecting and/or missing stuff.
        for mod_path, mod_name in _iter_modules(pkg_path):
            contrib = find_npe2_module_contributions(
                mod_path,
                plugin_name=plugin_name,
                module_name=f"{top_mod}.{mod_name}" if mod_name else top_mod,
            )
            contribs.append(contrib)

    mf = PluginManifest(
        name=plugin_name,
        contributions=merge_contributions(contribs),
    )

    if template is not None:
        mf = merge_manifests([template_mf, mf], overwrite=True)

    if dest is not None:
        manifest_string = getattr(mf, cast(str, suffix))(indent=2)
        pdest.write_text(manifest_string)

    return mf


def _iter_modules(path: Path) -> Iterator[Tuple[Path, str]]:
    """Return all python modules in path"""
    for p in path.glob("*.py"):
        yield p, "" if p.name == "__init__.py" else p.stem



import pytest

pytest_plugins = "pytester"

CASE1 = """
from npe2._pytest_plugin import TestPluginManager
from npe2 import PluginManager

def test_something_1(npe2pm):
    assert isinstance(npe2pm, TestPluginManager)
    assert PluginManager.instance() is npe2pm
"""

CASE2 = """
import pytest

def test_something_2(npe2pm, caplog):
    npe2pm.discover()
    assert "TestPluginManager refusing to discover plugins" in caplog.text
    assert len(caplog.records) == 1
    assert caplog.records[0].levelname == "WARNING"
"""

CASE3 = """
from npe2 import DynamicPlugin

def test_something_3(npe2pm):
    with npe2pm.tmp_plugin(name='some_name') as plugin:
        assert isinstance(plugin, DynamicPlugin)
        assert plugin.name in npe2pm._manifests
"""

CASE4 = """
from npe2 import PluginManifest

def test_something_4(npe2pm):
    mf = PluginManifest(name='some_name')
    with npe2pm.tmp_plugin(manifest=mf) as plugin:
        assert plugin.name in npe2pm._manifests
        assert plugin.manifest is mf
"""

CASE5 = """
import pytest
from importlib.metadata import PackageNotFoundError

def test_something_5(npe2pm):
    with pytest.raises(PackageNotFoundError):
        npe2pm.tmp_plugin(package='somepackage')
"""

CASE6 = """
import pytest

def test_something_6(npe2pm):
    with pytest.raises(FileNotFoundError):
        npe2pm.tmp_plugin(manifest='some_path.yaml')
"""


@pytest.mark.parametrize("case", [CASE1, CASE2, CASE3, CASE4, CASE5, CASE6])
def test_npe2pm_fixture(pytester: pytest.Pytester, case):
    """Make sure that the npe2pm fixture works."""

    # create a temporary pytest test file
    pytester.makepyfile(case)
    pytester.runpytest().assert_outcomes(passed=1)



from pathlib import Path

import pytest

from npe2._inspection import compile
from npe2.manifest.schema import PluginManifest


def test_compile(compiled_plugin_dir: Path, tmp_path: Path):
    """
    Test that the plugin manager can be compiled.
    """
    with pytest.raises(ValueError, match="must have an extension of .json, .yaml, or"):
        compile(compiled_plugin_dir, "bad_path")

    dest = tmp_path / "output.yaml"
    mf = compile(compiled_plugin_dir, dest=dest, packages=["my_module"])
    assert isinstance(mf, PluginManifest)
    assert mf.name == "my_compiled_plugin"
    assert mf.contributions.commands and len(mf.contributions.commands) == 5
    assert dest.exists()
    assert PluginManifest.from_file(dest) == mf


def test_compile_with_template(compiled_plugin_dir: Path, tmp_path: Path):
    """Test building from a template with npe2 compile."""
    template = tmp_path / "template.yaml"
    template.write_text("name: my_compiled_plugin\ndisplay_name: Display Name\n")
    mf = compile(compiled_plugin_dir, template=template)
    assert mf.name == "my_compiled_plugin"
    assert mf.display_name == "Display Name"



from __future__ import annotations

from typing import TYPE_CHECKING, Callable, Optional

from pydantic import Extra, Field

from npe2.manifest.utils import Executable
from npe2.types import Widget

if TYPE_CHECKING:
    from npe2._command_registry import CommandRegistry


class WidgetContribution(Executable[Widget]):
    """Contribute a widget that can be added to the napari viewer.

    Widget contributions point to a **command** that, when called, returns a widget
    *instance*; this includes functions that return a widget instance, (e.g. those
    decorated with `magicgui.magic_factory`) and subclasses of either
    [`QtWidgets.QWidget`](https://doc.qt.io/qt-5/qwidget.html) or
    [`magicgui.widgets.Widget`](https://napari.org/magicgui/api/_autosummary/magicgui.widgets._bases.Widget.html).

    Optionally, **autogenerate** may be used to create a widget (using
    [magicgui](https://napari.org/magicgui/)) from a command.  (In this case, the
    command needn't return a widget instance; it can be any function suitable as an
    argument to `magicgui.magicgui()`.)
    """

    command: str = Field(
        ...,
        description="Identifier of a command that returns a widget instance.  "
        "Or, if `autogenerate` is `True`, any command suitable as an argument "
        "to `magicgui.magicgui()`.",
    )
    display_name: str = Field(
        ..., description="Name for the widget, as presented in the UI."
    )
    autogenerate: bool = Field(
        default=False,
        description="If true, a widget will be autogenerated from the signature of "
        "the associated command using [magicgui](https://napari.org/magicgui/).",
    )

    class Config:
        extra = Extra.forbid

    def get_callable(
        self, _registry: Optional[CommandRegistry] = None
    ) -> Callable[..., Widget]:
        func = super().get_callable()
        if self.autogenerate:
            try:
                from magicgui import magic_factory
            except ImportError as e:
                raise ImportError(
                    "To use autogeneration, you must have magicgui installed."
                ) from e

            return magic_factory(func)
        return func



"""
# pyproject.toml
[build-system]
requires = ["setuptools", "wheel", "setuptools_scm", "npe2"]
build-backend = "setuptools.build_meta"

[tool.npe2]
"""
from __future__ import annotations

import os
import re
import sys
import warnings
from typing import TYPE_CHECKING, Optional, Tuple, cast

from setuptools import Distribution
from setuptools.command.build_py import build_py

if TYPE_CHECKING:
    from distutils.cmd import Command
    from typing import Any, Union

    PathT = Union["os.PathLike[str]", str]

NPE2_ENTRY = "napari.manifest"
DEBUG = bool(os.environ.get("SETUPTOOLS_NPE2_DEBUG"))
EP_PATTERN = re.compile(
    r"(?P<module>[\w.]+)\s*(:\s*(?P<attr>[\w.]+)\s*)?((?P<extras>\[.*\])\s*)?$"
)


def trace(*k: object) -> None:
    if DEBUG:
        print(*k, file=sys.stderr, flush=True)


def _lazy_tomli_load(data: str) -> dict[str, Any]:
    from pytomlpp import loads

    return loads(data)


def _read_dist_name_from_setup_cfg() -> str | None:
    # minimal effort to read dist_name off setup.cfg metadata
    import configparser

    parser = configparser.ConfigParser()
    parser.read(["setup.cfg"])
    return parser.get("metadata", "name", fallback=None)


def _check_absolute_root(root: PathT, relative_to: PathT | None) -> str:
    trace("abs root", repr(locals()))
    if relative_to:
        if (
            os.path.isabs(root)
            and os.path.commonpath([root, relative_to]) != relative_to
        ):
            warnings.warn(
                f"absolute root path '{root}' overrides relative_to '{relative_to}'",
                stacklevel=2,
            )
        if os.path.isdir(relative_to):
            warnings.warn(
                "relative_to is expected to be a file,"
                " its the directory {relative_to!r}\n"
                "assuming the parent directory was passed",
                stacklevel=2,
            )
            trace("dir", relative_to)
            root = os.path.join(relative_to, root)
        else:
            trace("file", relative_to)
            root = os.path.join(os.path.dirname(relative_to), root)
    return os.path.abspath(root)


class Configuration:
    """Global configuration model"""

    def __init__(
        self,
        relative_to: PathT | None = None,
        root: PathT = ".",
        write_to: PathT | None = None,
        write_to_template: str | None = None,
        dist_name: str | None = None,
        template: str | None = None,
    ):
        self._relative_to = None if relative_to is None else os.fspath(relative_to)
        self._root = "."
        self.root = os.fspath(root)
        self.write_to = write_to
        self.write_to_template = write_to_template
        self.dist_name = dist_name
        self.template = template

    @property
    def relative_to(self) -> str | None:
        return self._relative_to

    @property
    def root(self) -> str:
        return self._root

    @root.setter
    def root(self, value: PathT) -> None:
        self._absolute_root = _check_absolute_root(value, self._relative_to)
        self._root = os.fspath(value)
        trace("root", repr(self._absolute_root))
        trace("relative_to", repr(self._relative_to))

    @property
    def absolute_root(self) -> str:
        return self._absolute_root

    @classmethod
    def from_file(
        cls, name: str = "pyproject.toml", dist_name: str | None = None, **kwargs: Any
    ) -> Configuration:
        """
        Read Configuration from pyproject.toml (or similar).
        Raises exceptions when file is not found or toml is
        not installed or the file has invalid format or does
        not contain the [tool.npe2] section.
        """

        with open(name, encoding="UTF-8") as strm:
            data = strm.read()
        defn = _lazy_tomli_load(data)
        try:
            section = defn.get("tool", {})["npe2"]
        except LookupError as e:
            raise LookupError(f"{name} does not contain a tool.npe2 section") from e
        if "dist_name" in section:
            if dist_name is None:
                dist_name = section.pop("dist_name")
            else:
                assert dist_name == section["dist_name"]
                del section["dist_name"]
        if dist_name is None and "project" in defn:
            # minimal pep 621 support for figuring the pretend keys
            dist_name = defn["project"].get("name")
        if dist_name is None:
            dist_name = _read_dist_name_from_setup_cfg()

        return cls(dist_name=dist_name, **section, **kwargs)


def _mf_entry_from_dist(dist: Distribution) -> Optional[Tuple[str, str]]:
    """Return (module, attr) for a distribution's npe2 entry point."""
    eps: dict = getattr(dist, "entry_points", {})
    if napari_entrys := eps.get(NPE2_ENTRY, []):
        if match := EP_PATTERN.search(napari_entrys[0]):
            return match.group("module"), match.group("attr")
    return None


class npe2_compile(build_py):
    def run(self) -> None:
        trace("RUN npe2_compile")
        if ep := _mf_entry_from_dist(self.distribution):
            from npe2._inspection._compile import compile

            module, attr = ep
            src = self.distribution.src_root or os.getcwd()
            dest = os.path.join(self.get_package_dir(module), attr)
            compile(src, dest, template=self.distribution.config.template)
        else:
            name = self.distribution.metadata.name
            trace(f"no {NPE2_ENTRY!r} found in entry_points for {name}")


def finalize_npe2(dist: Distribution):
    # this hook is declared in the setuptools.finalize_distribution_options
    # entry point in our setup.cfg
    # https://setuptools.pypa.io/en/latest/userguide/extension.html#customizing-distribution-options
    trace("finalize hook", vars(dist.metadata))
    dist_name = dist.metadata.name
    if dist_name is None:
        dist_name = _read_dist_name_from_setup_cfg()
    if not os.path.isfile("pyproject.toml"):
        return
    if dist_name == "npe2":
        # if we're packaging npe2 itself, don't do anything
        return
    try:
        # config will *only* be detected if there is a [tool.npe2]
        # section in pyproject.toml.  This is how plugins opt in
        # to the npe2 compile feature during build
        config = Configuration.from_file(dist_name=dist_name)
    except LookupError as e:
        trace(e)
    else:
        # inject our `npe2_compile` command to be called whenever we're building an
        # sdist or a wheel
        dist.config = config
        for cmd in ("build", "sdist"):
            if base := dist.get_command_class(cmd):
                cast("Command", base).sub_commands.insert(0, ("npe2_compile", None))



# mypy: disable-error-code=empty-body
"""Convenience module to access methods on the global PluginManager singleton."""
from __future__ import annotations

from typing import TYPE_CHECKING, Dict

if TYPE_CHECKING:
    from os import PathLike
    from typing import Any, Iterator, List, NewType, Optional, Sequence, Tuple, Union

    from npe2 import PluginManifest
    from npe2._plugin_manager import InclusionSet, PluginContext
    from npe2.manifest import contributions

    from ._plugin_manager import PluginManager

    PluginName = NewType("PluginName", str)


def instance() -> PluginManager:
    """Return global PluginManager singleton instance."""
    from ._plugin_manager import PluginManager

    return PluginManager.instance()


def discover(paths: Sequence[str] = (), clear=False, include_npe1=False) -> None:
    """Discover and index plugin manifests in the environment."""


def dict(
    self,
    *,
    include: Optional[InclusionSet] = None,
    exclude: Optional[InclusionSet] = None,
) -> Dict[str, Any]:
    """Return a dictionary with the state of the plugin manager."""


def index_npe1_adapters() -> None:
    """Import and index any/all npe1 adapters."""


def register(manifest: PluginManifest, warn_disabled=True) -> None:
    """Register a plugin manifest"""


def unregister(key: PluginName) -> None:
    """Unregister plugin named `key`."""


def activate(key: PluginName) -> PluginContext:
    """Activate plugin with `key`."""


def get_context(plugin_name: PluginName) -> PluginContext:
    """Return PluginContext for plugin_name"""


def deactivate(plugin_name: PluginName) -> None:
    """Deactivate `plugin_name`"""


def enable(plugin_name: PluginName) -> None:
    """Enable a plugin (which mostly means just `un-disable` it."""


def disable(plugin_name: PluginName) -> None:
    """Disable a plugin"""


def is_disabled(plugin_name: str) -> bool:
    """Return `True` if plugin_name is disabled."""


def get_manifest(plugin_name: str) -> PluginManifest:
    """Get manifest for `plugin_name`"""


def iter_manifests(disabled: Optional[bool] = None) -> Iterator[PluginManifest]:
    """Iterate through registered manifests."""


def get_command(command_id: str) -> contributions.CommandContribution:
    """Retrieve CommandContribution for `command_id`"""


def get_submenu(submenu_id: str) -> contributions.SubmenuContribution:
    """Get SubmenuContribution for `submenu_id`."""


def iter_menu(menu_key: str, disabled=False) -> Iterator[contributions.MenuItem]:
    """Iterate over `MenuItems` in menu with id `menu_key`."""


def menus(disabled=False) -> Dict[str, List[contributions.MenuItem]]:
    """Return all registered menu_key -> List[MenuItems]."""


def iter_themes() -> Iterator[contributions.ThemeContribution]:
    """Iterate over discovered/enuabled `ThemeContributions`."""


def iter_compatible_readers(
    path: Union[PathLike, Sequence[str]]
) -> Iterator[contributions.ReaderContribution]:
    """Iterate over ReaderContributions compatible with `path`."""


def iter_compatible_writers(
    layer_types: Sequence[str],
) -> Iterator[contributions.WriterContribution]:
    """Iterate over compatible WriterContributions given a sequence of layer_types."""


def iter_widgets() -> Iterator[contributions.WidgetContribution]:
    """Iterate over discovered WidgetContributions."""


def iter_sample_data() -> (
    Iterator[Tuple[PluginName, List[contributions.SampleDataContribution]]]
):
    """Iterates over (plugin_name, [sample_contribs])."""


def get_writer(
    path: str, layer_types: Sequence[str], plugin_name: Optional[str] = None
) -> Tuple[Optional[contributions.WriterContribution], str]:
    """Get Writer contribution appropriate for `path`, and `layer_types`."""


def _populate_module():
    """Convert all functions in this module into global plugin manager methods."""
    import functools
    import sys

    from ._plugin_manager import PluginManager

    _module = sys.modules[__name__]
    for key in dir(_module):
        if key.startswith(("_", "instance")) or not hasattr(PluginManager, key):
            continue

        @functools.wraps(getattr(_module, key))
        def _f(*args, _key=key, **kwargs):
            return getattr(instance(), _key)(*args, **kwargs)

        setattr(_module, key, _f)


_populate_module()
del _populate_module, TYPE_CHECKING, annotations



import shutil
import sys
from importlib import abc, metadata
from pathlib import Path
from unittest.mock import patch

import pytest

from npe2 import PluginManager, PluginManifest
from npe2.manifest import _npe1_adapter

FIXTURES = Path(__file__).parent / "fixtures"


@pytest.fixture
def sample_path():
    return Path(__file__).parent / "sample"


@pytest.fixture
def sample_manifest(sample_path):
    return PluginManifest.from_file(sample_path / "my_plugin" / "napari.yaml")


@pytest.fixture
def compiled_plugin_dir(tmp_path):
    shutil.copytree(FIXTURES / "my-compiled-plugin", tmp_path, dirs_exist_ok=True)
    return tmp_path


@pytest.fixture
def uses_sample_plugin(sample_path):
    sys.path.append(str(sample_path))
    try:
        pm = PluginManager.instance()
        pm.discover()
        yield
    finally:
        sys.path.remove(str(sample_path))


@pytest.fixture
def plugin_manager():
    pm = PluginManager()
    pm.discover()
    return pm


@pytest.fixture(autouse=True)
def mock_discover():
    _discover = PluginManifest.discover

    def wrapper(*args, **kwargs):
        before = sys.path.copy()
        # only allowing things from test directory in discover
        sys.path = [x for x in sys.path if str(Path(__file__).parent) in x]
        try:
            yield from _discover(*args, **kwargs)
        finally:
            sys.path = before

    with patch("npe2.PluginManifest.discover", wraps=wrapper):
        yield 1


@pytest.fixture
def npe1_repo():
    return Path(__file__).parent / "npe1-plugin"


@pytest.fixture
def uses_npe1_plugin(npe1_repo):
    import site

    class Importer(abc.MetaPathFinder):
        def find_spec(self, *_, **__):
            return None

        def find_distributions(self, ctx, **k):
            if ctx.name == "npe1-plugin":
                pth = npe1_repo / "npe1-plugin-0.0.1.dist-info"
                yield metadata.PathDistribution(pth)
            return

    sys.meta_path.append(Importer())
    sys.path.append(str(npe1_repo))
    try:
        pkgs = [*site.getsitepackages(), str(npe1_repo)]
        with patch("site.getsitepackages", return_value=pkgs):
            yield
    finally:
        sys.path.remove(str(npe1_repo))


@pytest.fixture
def npe1_plugin_module(npe1_repo):
    import sys
    from importlib.util import module_from_spec, spec_from_file_location

    npe1_module_path = npe1_repo / "npe1_module" / "__init__.py"
    spec = spec_from_file_location("npe1_module", npe1_module_path)
    assert spec
    module = module_from_spec(spec)
    sys.modules[spec.name] = module
    spec.loader.exec_module(module)  # type: ignore
    try:
        yield module
    finally:
        del sys.modules[spec.name]


@pytest.fixture
def mock_npe1_pm():
    from napari_plugin_engine import PluginManager, napari_hook_specification

    # fmt: off
    class HookSpecs:
        def napari_provide_sample_data(): ...  # type: ignore
        def napari_get_reader(path): ...
        def napari_get_writer(path, layer_types): ...
        def napari_write_image(path, data, meta): ...
        def napari_write_labels(path, data, meta): ...
        def napari_write_points(path, data, meta): ...
        def napari_write_shapes(path, data, meta): ...
        def napari_write_surface(path, data, meta): ...
        def napari_write_vectors(path, data, meta): ...
        def napari_experimental_provide_function(): ...  # type: ignore
        def napari_experimental_provide_dock_widget(): ...  # type: ignore
        def napari_experimental_provide_theme(): ...  # type: ignore
    # fmt: on

    for m in dir(HookSpecs):
        if m.startswith("napari"):
            setattr(HookSpecs, m, napari_hook_specification(getattr(HookSpecs, m)))

    pm = PluginManager("napari")
    pm.add_hookspecs(HookSpecs)

    yield pm


@pytest.fixture
def mock_npe1_pm_with_plugin(npe1_repo, npe1_plugin_module):
    """Mocks a fully installed local repository"""
    from npe2._inspection._from_npe1 import plugin_packages

    mock_dist = metadata.PathDistribution(npe1_repo / "npe1-plugin-0.0.1.dist-info")

    def _dists():
        return [mock_dist]

    def _from_name(name):
        if name == "npe1-plugin":
            return mock_dist
        raise metadata.PackageNotFoundError(name)

    setup_cfg = npe1_repo / "setup.cfg"
    new_manifest = npe1_repo / "npe1_module" / "napari.yaml"
    with patch.object(metadata, "distributions", new=_dists):
        with patch.object(metadata.Distribution, "from_name", new=_from_name):
            cfg = setup_cfg.read_text()
            plugin_packages.cache_clear()
            try:
                yield mock_npe1_pm
            finally:
                plugin_packages.cache_clear()
                setup_cfg.write_text(cfg)
                if new_manifest.exists():
                    new_manifest.unlink()
                if (npe1_repo / "setup.py").exists():
                    (npe1_repo / "setup.py").unlink()


@pytest.fixture(autouse=True)
def mock_cache(tmp_path, monkeypatch):
    with monkeypatch.context() as m:
        m.setattr(_npe1_adapter, "ADAPTER_CACHE", tmp_path)
        yield tmp_path



from npe2 import PluginManager


def test_pm_module():
    from npe2 import plugin_manager as pm

    assert pm.instance() is PluginManager.instance()

    # smoke-test checking that a few of the argument-free things work
    # they may or may-not be empty depending on other tests in this suite.
    pm.iter_widgets()
    pm.iter_sample_data()

    # make sure we have it covered.
    for k, v in vars(PluginManager).items():
        if k.startswith("_") or isinstance(v, (classmethod, property)):
            continue
        assert hasattr(pm, k), f"pm.py module is missing function {k!r}"



from typing import Dict, List, Optional

from pydantic import BaseModel, Field, validator

from ._commands import CommandContribution
from ._configuration import ConfigurationContribution
from ._keybindings import KeyBindingContribution
from ._menus import MenuItem
from ._readers import ReaderContribution
from ._sample_data import SampleDataContribution, SampleDataGenerator, SampleDataURI
from ._submenu import SubmenuContribution
from ._themes import ThemeContribution
from ._widgets import WidgetContribution
from ._writers import WriterContribution

__all__ = [
    "ContributionPoints",
    "CommandContribution",
    "KeyBindingContribution",
    "MenuItem",
    "ReaderContribution",
    "SampleDataContribution",
    "SubmenuContribution",
    "ThemeContribution",
    "WidgetContribution",
    "WriterContribution",
    "SampleDataGenerator",
    "SampleDataURI",
]


class ContributionPoints(BaseModel):
    commands: Optional[List[CommandContribution]]
    readers: Optional[List[ReaderContribution]]
    writers: Optional[List[WriterContribution]]
    widgets: Optional[List[WidgetContribution]]
    sample_data: Optional[List[SampleDataContribution]]
    themes: Optional[List[ThemeContribution]]
    menus: Dict[str, List[MenuItem]] = Field(default_factory=dict, hide_docs=True)
    submenus: Optional[List[SubmenuContribution]] = Field(None, hide_docs=True)
    keybindings: Optional[List[KeyBindingContribution]] = Field(None, hide_docs=True)

    configuration: List[ConfigurationContribution] = Field(
        default_factory=list,
        hide_docs=True,
        description="Configuration options for this plugin."
        "This section can either be a single object, representing a single category of"
        "settings, or an array of objects, representing multiple categories of"
        "settings. If there are multiple categories of settings, the Settings editor"
        "will show a submenu in the table of contents for that extension, and the title"
        "keys will be used for the submenu entry names.",
    )

    @validator("configuration", pre=True)
    def _to_list(cls, v):
        return v if isinstance(v, list) else [v]



from __future__ import annotations

import io
import json
import os
import re
import subprocess
import tempfile
from concurrent.futures import ProcessPoolExecutor
from contextlib import contextmanager
from functools import lru_cache
from importlib import metadata
from logging import getLogger
from pathlib import Path
from typing import (
    TYPE_CHECKING,
    Any,
    ContextManager,
    Dict,
    Iterator,
    List,
    Optional,
    Tuple,
    Union,
)
from unittest.mock import patch
from urllib import error, parse, request
from zipfile import ZipFile

from npe2.manifest import PackageMetadata

if TYPE_CHECKING:
    from npe2.manifest import PluginManifest


logger = getLogger(__name__)

NPE1_ENTRY_POINT = "napari.plugin"
NPE2_ENTRY_POINT = "napari.manifest"
__all__ = [
    "fetch_manifest",
    "get_pypi_url",
    "get_hub_plugin",
    "get_hub_plugins",
    "get_pypi_plugins",
]


def _manifest_from_npe2_dist(
    dist: metadata.PathDistribution, ep: metadata.EntryPoint
) -> PluginManifest:
    """Extract plugin manifest from a distribution with an npe2 entry point."""
    from npe2.manifest import PluginManifest

    logger.debug("pypi wheel has npe2 entry point.")

    # python 3.8 fallbacks
    match = ep.pattern.match(ep.value)
    assert match
    module: str = match.groupdict()["module"]
    attr: str = match.groupdict()["attr"]

    mf_file = Path(dist.locate_file(Path(module) / attr))
    if not mf_file.exists():
        raise ValueError(  # pragma: no cover
            f"manifest {mf_file.name!r} does not exist in distribution "
            f"for {dist.metadata['Name']}"
        )

    mf = PluginManifest.from_file(str(mf_file))
    # manually add the package metadata from our distribution object.
    mf.package_metadata = PackageMetadata.from_dist_metadata(dist.metadata)
    return mf


def _manifest_from_npe1_dist(dist: metadata.PathDistribution) -> PluginManifest:
    """Extract plugin manifest from a distribution with an npe1 entry point."""
    from npe2.manifest import PluginManifest
    from npe2.manifest.utils import merge_contributions

    from . import find_npe1_module_contributions

    name = dist.metadata["Name"]
    contribs = []
    for ep in dist.entry_points:
        if ep.group == NPE1_ENTRY_POINT and (match := ep.pattern.match(ep.value)):
            module = match.group("module")
            contribs.append(find_npe1_module_contributions(dist, module))

    mf = PluginManifest(
        name=name, contributions=merge_contributions(contribs), npe1_shim=True
    )
    mf.package_metadata = PackageMetadata.from_dist_metadata(dist.metadata)
    return mf


def _manifest_from_extracted_wheel(wheel_dir: Path) -> PluginManifest:
    """Return plugin manifest from an extracted wheel."""
    # create a PathDistribution from the dist-info directory in the wheel
    dist = metadata.PathDistribution(next(Path(wheel_dir).glob("*.dist-info")))

    has_npe1 = False
    for ep in dist.entry_points:
        # if we find an npe2 entry point, we can just use
        # PathDistribution.locate_file to get the file.
        if ep.group == NPE2_ENTRY_POINT:
            return _manifest_from_npe2_dist(dist, ep)
        elif ep.group == NPE1_ENTRY_POINT:
            has_npe1 = True  # pragma: no cover
    if has_npe1:
        return _manifest_from_npe1_dist(dist)

    raise ValueError("No npe2 or npe1 entry point found in wheel")  # pragma: no cover


@contextmanager
def _guard_cwd() -> Iterator[None]:
    """Protect current working directory from changes."""
    current = os.getcwd()
    try:
        yield
    finally:
        os.chdir(current)


def _build_wheel(src: Union[str, Path]) -> Path:
    """Build a wheel from a source directory and extract it into dest."""
    from build.__main__ import build_package

    dest = Path(src) / "extracted_wheel"

    class _QuietPopen(subprocess.Popen):
        """Silence all the noise from build."""

        def __init__(self, *args, **kwargs):
            kwargs["stdout"] = subprocess.DEVNULL
            kwargs["stderr"] = subprocess.DEVNULL
            super().__init__(*args, **kwargs)

    with patch("subprocess.Popen", _QuietPopen), _guard_cwd():
        dist = Path(src) / "dist"
        build_package(src, dist, ["wheel"])
        with ZipFile(next((dist).glob("*.whl"))) as zf:
            zf.extractall(dest)
    return dest


def get_manifest_from_wheel(src: str) -> PluginManifest:
    """Extract a manifest from a .whl file."""
    with tempfile.TemporaryDirectory() as td:
        with ZipFile(src) as zf:
            zf.extractall(td)
            return _manifest_from_extracted_wheel(Path(td))


def _build_src_and_extract_manifest(src_dir: Union[str, Path]) -> PluginManifest:
    """Build a wheel from a source directory and extract the manifest."""
    return _manifest_from_extracted_wheel(_build_wheel(src_dir))


def _get_manifest_from_zip_url(url: str) -> PluginManifest:
    """Extract a manifest from a remote source directory zip file.

    Examples
    --------
    $ npe2 fetch https://github.com/org/project/archive/refs/heads/master.zip
    """
    with _tmp_zip_download(url) as zip_path:
        src_dir = next(Path(zip_path).iterdir())  # find first directory
        return _build_src_and_extract_manifest(src_dir)


def _get_manifest_from_wheel_url(url: str) -> PluginManifest:
    """Extract a manifest from a remote wheel file.

    Examples
    --------
    $ npe2 fetch https://files.pythonhosted.org/packages/b0/93/a00a1ee154d5ce3540dd5ae081dc53fcfa7498f34ba68a7345ac027a4f96/pycudadecon-0.3.0-py3-none-any.whl
    """
    with _tmp_zip_download(url) as wheel_dir:
        return _manifest_from_extracted_wheel(wheel_dir)


def _get_manifest_from_targz_url(url: str) -> PluginManifest:
    """Extract a manifest from a remote source directory tar.gz file.

    Examples
    --------
    $ npe2 fetch https://files.pythonhosted.org/packages/4a/84/de031ba465f183c319cb37633c49dfebd57f1ff42bc9744db3f80f7f4093/pycudadecon-0.3.0.tar.gz
    """
    with _tmp_targz_download(url) as targz_path:
        src_dir = next(Path(targz_path).iterdir())  # find first directory
        return _build_src_and_extract_manifest(src_dir)


def _get_manifest_from_git_url(url: str) -> PluginManifest:
    """Extract a manifest from a remote git repository.

    Examples
    --------
    $ npe2 fetch https://github.com/tlambert03/napari-dv
    $ npe2 fetch https://github.com/tlambert03/napari-dv.git
    $ npe2 fetch git+https://github.com/tlambert03/napari-dv.git
    """
    if url.startswith("git+"):
        url = url[4:]

    with tempfile.TemporaryDirectory() as td:
        subprocess.run(["git", "clone", url, td], stdout=subprocess.DEVNULL)
        return _build_src_and_extract_manifest(td)


def fetch_manifest(
    package_or_url: str, version: Optional[str] = None
) -> PluginManifest:
    """Fetch a manifest for a pypi package name or URL to a wheel or source.

    Parameters
    ----------
    package_or_url : str
        package name or URL to a git repository or zip file.
    version : Optional[str]
        package version, by default, latest version.

    Returns
    -------
    PluginManifest
        Plugin manifest for package `specifier`.

    Examples
    --------
    >>> fetch_manifest("napari-dv")
    >>> fetch_manifest("napari-dv", "0.3.0")
    >>> fetch_manifest("https://github.com/tlambert03/napari-dv")
    >>> fetch_manifest("git+https://github.com/tlambert03/napari-dv.git")
    >>> fetch_manifest("https://github.com/org/project/archive/refs/heads/master.zip")
    >>> fetch_manifest("https://files.pythonhosted.org/.../package-0.3.0-py3-none-any.whl")
    >>> fetch_manifest("https://files.pythonhosted.org/.../package-0.3.0.tar.gz")
    """
    # not on PyPI check various URL forms
    if package_or_url.startswith(("http", "git+http")):
        if package_or_url.endswith(".zip"):
            return _get_manifest_from_zip_url(package_or_url)
        if package_or_url.endswith(".whl"):
            return _get_manifest_from_wheel_url(package_or_url)
        if package_or_url.endswith(".tar.gz"):
            return _get_manifest_from_targz_url(package_or_url)
        if (
            package_or_url.startswith("git+")
            or package_or_url.endswith(".git")
            or "github.com" in package_or_url
        ):
            return _get_manifest_from_git_url(package_or_url)
    else:
        try:
            with _tmp_pypi_wheel_download(package_or_url, version) as td:
                return _manifest_from_extracted_wheel(td)
        except metadata.PackageNotFoundError:
            return _manifest_from_pypi_sdist(package_or_url, version)
        except error.HTTPError:
            pass  # pragma: no cover
    raise ValueError(  # pragma: no cover
        f"Could not interpret {package_or_url!r} as a PYPI package name or URL to a "
        "wheel or source distribution/zip file."
    )


def _manifest_from_pypi_sdist(
    package: str, version: Optional[str] = None
) -> PluginManifest:
    """Extract a manifest from a source distribution on pypi."""
    with _tmp_pypi_sdist_download(package, version) as td:
        src = next(p for p in td.iterdir() if p.is_dir())
        return _build_src_and_extract_manifest(src)


@lru_cache
def _pypi_info(package: str) -> dict:
    with request.urlopen(f"https://pypi.org/pypi/{package}/json") as f:
        return json.load(f)


def get_pypi_url(
    package: str, version: Optional[str] = None, packagetype: Optional[str] = None
) -> str:
    """Get URL for a package on PyPI.

    Parameters
    ----------
    package : str
        package name
    version : str, optional
        package version, by default, latest version.
    packagetype : str, optional
        one of `'sdist'`, `'bdist_wheel'`, or `None`,
        by default `None`, which means 'bdist_wheel' will be tried first, then 'sdist'

    Returns
    -------
    str
        URL to download the package.

    Raises
    ------
    ValueError
        If packagetype is not one of 'sdist', 'bdist_wheel', or if version is specified
        and does not match any available version.
    KeyError
        If packagetype is specified and no package of that type is available.
    """
    if packagetype not in {"sdist", "bdist_wheel", None}:
        raise ValueError(  # pragma: no cover
            f"Invalid packagetype: {packagetype}, must be one of sdist, bdist_wheel"
        )

    data = _pypi_info(package)
    if version:
        version = version.lstrip("v")
        try:
            _releases: List[dict] = data["releases"][version]
        except KeyError as e:  # pragma: no cover
            raise ValueError(f"{package} does not have version {version}") from e
    else:
        _releases = data["urls"]

    releases = {d.get("packagetype"): d for d in _releases}
    if packagetype:
        if packagetype not in releases:  # pragma: no cover
            version = version or "latest"
            raise metadata.PackageNotFoundError(
                f'No {packagetype} releases found for version "{version}"'
            )
        return releases[packagetype]["url"]
    return (releases.get("bdist_wheel") or releases["sdist"])["url"]


@contextmanager
def _tmp_zip_download(url: str) -> Iterator[Path]:
    """Extract remote zip file to a temporary directory."""
    with tempfile.TemporaryDirectory() as td, request.urlopen(url) as f:
        with ZipFile(io.BytesIO(f.read())) as zf:
            zf.extractall(td)
            yield Path(td)


@contextmanager
def _tmp_targz_download(url: str) -> Iterator[Path]:
    """Extract remote tar.gz file to a temporary directory."""
    import tarfile

    with tempfile.TemporaryDirectory() as td, request.urlopen(url) as f:
        with tarfile.open(fileobj=f, mode="r:gz") as tar:
            tar.extractall(td)
            yield Path(td)


def _tmp_pypi_wheel_download(
    package: str, version: Optional[str] = None
) -> ContextManager[Path]:
    url = get_pypi_url(package, version=version, packagetype="bdist_wheel")
    logger.debug(f"downloading wheel for {package} {version or ''}")
    return _tmp_zip_download(url)


def _tmp_pypi_sdist_download(
    package: str, version: Optional[str] = None
) -> ContextManager[Path]:
    url = get_pypi_url(package, version=version, packagetype="sdist")
    logger.debug(f"downloading sdist for {package} {version or ''}")
    return _tmp_targz_download(url)


@lru_cache
def _get_packages_by_classifier(classifier: str) -> Dict[str, str]:
    """Search for packages declaring ``classifier`` on PyPI.

    Returns
    -------
    packages : List[str]
        name of all packages at pypi that declare ``classifier``
    """
    PACKAGE_NAME_PATTERN = re.compile('class="package-snippet__name">(.+)</span>')
    PACKAGE_VERSION_PATTERN = re.compile('class="package-snippet__version">(.+)</span>')

    packages = {}
    page = 1
    url = f"https://pypi.org/search/?c={parse.quote_plus(classifier)}&page="
    while True:
        try:
            with request.urlopen(f"{url}{page}") as response:
                html = response.read().decode()
            names = PACKAGE_NAME_PATTERN.findall(html)
            versions = PACKAGE_VERSION_PATTERN.findall(html)
            packages.update(dict(zip(names, versions)))
            page += 1
        except error.HTTPError:
            break

    return dict(sorted(packages.items()))


def get_pypi_plugins() -> Dict[str, str]:
    """Return {name: latest_version} for all plugins found on pypi."""
    NAPARI_CLASSIFIER = "Framework :: napari"
    return _get_packages_by_classifier(NAPARI_CLASSIFIER)


@lru_cache
def get_hub_plugins() -> Dict[str, str]:
    """Return {name: latest_version} for all plugins on the hub."""
    with request.urlopen("https://api.napari-hub.org/plugins") as r:
        return json.load(r)


@lru_cache
def get_hub_plugin(plugin_name: str) -> Dict[str, Any]:
    """Return hub information for a specific plugin."""
    with request.urlopen(f"https://api.napari-hub.org/plugins/{plugin_name}") as r:
        return json.load(r)


def _try_fetch_and_write_manifest(args: Tuple[str, str, Path, int]):
    name, version, dest, indent = args
    FORMAT = "json"

    try:  # pragma: no cover
        mf = fetch_manifest(name, version=version)
        manifest_string = getattr(mf, FORMAT)(exclude=set(), indent=indent)

        (dest / f"{name}.{FORMAT}").write_text(manifest_string)
        print(f"✅ {name}")
    except Exception as e:
        print(f"❌ {name}")
        return name, {"version": version, "error": str(e)}


def fetch_all_manifests(dest: str = "manifests", indent: int = 2) -> None:
    """Fetch all manifests for plugins on PyPI and write to ``dest`` directory."""
    _dest = Path(dest)
    _dest.mkdir(exist_ok=True, parents=True)

    args = [
        (name, ver, _dest, indent) for name, ver in sorted(get_pypi_plugins().items())
    ]

    # use processes instead of threads, because many of the subroutines in build
    # and setuptools use `os.chdir()`, which is not thread-safe
    with ProcessPoolExecutor() as executor:
        errors = list(executor.map(_try_fetch_and_write_manifest, args))
    _errors = {tup[0]: tup[1] for tup in errors if tup}
    (_dest / "errors.json").write_text(json.dumps(_errors, indent=indent))



from __future__ import annotations

import sys
from contextlib import contextmanager, suppress
from enum import Enum
from importlib import metadata, util
from logging import getLogger
from pathlib import Path
from typing import Iterator, List, Literal, NamedTuple, Optional, Sequence, Union

from pydantic import Extra, Field, ValidationError, root_validator, validator
from pydantic.error_wrappers import ErrorWrapper
from pydantic.main import BaseModel, ModelMetaclass

from npe2.types import PythonName

from . import _validators
from ._bases import ImportExportModel
from ._package_metadata import PackageMetadata
from .contributions import ContributionPoints
from .utils import Executable, Version

logger = getLogger(__name__)


SCHEMA_VERSION = "0.2.0"
ENTRY_POINT = "napari.manifest"
NPE1_ENTRY_POINT = "napari.plugin"


class Category(str, Enum):
    """Broad plugin categories, values for PluginManifest.categories."""

    # drive devices (webcams, microscopes, etc) to acquire data
    Acquisition = "Acquisition"
    # tools that facilitate labeling, marking, and, erm, "annotating" data within napari
    Annotation = "Annotation"
    # Sample datasets for training, demonstration, learning, etc...
    Dataset = "Dataset"
    # Routines that take in numerical arrays and generally return new arrays or datasets
    # (e.g. scikit image stuff, deconvolution, super-res reconstruction, etc...)
    Image_Processing = "Image Processing"
    # Plugins that read from and/or write to files or data streams
    # not supported natively by napari
    IO = "IO"
    # Plugins that employ machine learning: may facilitate either training or prediction
    # Machine_Learning = "Machine Learning"

    # Tools that extract measurements (i.e. into tabular, graph, or other data formats),
    # such as region properties, etc...
    Measurement = "Measurement"
    # tools that identify objects and/or boundaries in datasets
    # (including, but not limited to, images)
    Segmentation = "Segmentation"
    # tools that simulate some physical process.
    # microscope/PSF generators, optics simulators, astronomy simulations, etc...
    Simulation = "Simulation"
    # plugins that modify the look and feel of napari
    Themes = "Themes"
    # linear and or nonlinear registrations, general data transformations and mappings
    Transformations = "Transformations"
    # Conveniences, widgets, etc... stuff that could conceivably be "core"
    # but which is community-supported
    Utilities = "Utilities"
    # tools for plotting, rendering, and visualization
    # (on top of those provided by napari)
    Visualization = "Visualization"

    def __str__(self) -> str:
        return self.value  # pragma: no cover


class DiscoverResults(NamedTuple):
    manifest: Optional[PluginManifest]
    distribution: Optional[metadata.Distribution]
    error: Optional[Exception]


class PluginManifest(ImportExportModel):
    class Config:
        underscore_attrs_are_private = True
        extra = Extra.ignore
        validate_assignment = True

    # VS Code uses <publisher>.<name> as a unique ID for the extension
    # should this just be the package name ... not the module name? (yes)
    # do we normalize this? (i.e. underscores / dashes ?) (no)
    # TODO: enforce that this matches the package name

    name: str = Field(
        ...,
        description="The name of the plugin. Though this field is mandatory, it *must*"
        " match the package `name` as defined in the python package metadata.",
        allow_mutation=False,
    )
    _validate_name = validator("name", pre=True, allow_reuse=True)(
        _validators.package_name
    )

    display_name: str = Field(
        "",
        description="User-facing text to display as the name of this plugin. "
        "Must be 3-90 characters long and must not begin or end with an underscore, "
        "white space, or non-word character. If not provided, the manifest `name` "
        "will be used as the display name.",
        min_length=3,
        max_length=90,
    )

    _validate_display_name = validator("display_name", allow_reuse=True)(
        _validators.display_name
    )

    visibility: Literal["public", "hidden"] = Field(
        "public",
        description="Whether this plugin should be searchable and visible in "
        "the built-in plugin installer and the napari hub. By default (`'public'`) "
        "all plugins are visible. To prevent your plugin from appearing in search "
        "results, change this to `'hidden'`.",
    )

    icon: str = Field(
        "",
        description="The path to a square PNG icon of at least 128x128 pixels (256x256 "
        "for Retina screens). May be one of:\n"
        "  - a secure (https) URL\n"
        "  - a path relative to the manifest file, (must be shipped in the sdist)\n"
        "  - a string in the format `{package}:{resource}`, where `package` and "
        "`resource` are arguments to `importlib.resources.path(package, resource)`, "
        "(e.g. `top_module.some_folder:my_logo.png`).",
    )
    _validate_icon_path = validator("icon", allow_reuse=True)(_validators.icon_path)

    categories: List[Category] = Field(
        default_factory=list,
        description="A list of categories that this plugin belongs to. This is used to "
        "help users discover your plugin. Allowed values:\n"
        f"`[{', '.join(Category)}]`",
    )
    # Plugins rely on certain guarantees to interoperate propertly with the
    # plugin engine. These include the manifest specification, conventions
    # around python packaging, command api's, etc. Together these form a
    # "contract". The version of this contract is the "schema version."
    #
    # The first release of npe2 defines the first schema version.
    # As the contract around plugins evolve the SCHEMA_VERSION should be
    # increased follow SemVer rules. Note that sometimes the version number
    # will change even though no npe2 code changes.
    #
    # The `schema_version` field declares the version of the contract that this
    # plugin targets.
    schema_version: str = Field(
        SCHEMA_VERSION,
        description="A SemVer compatible version string matching the napari plugin "
        "schema version that the plugin is compatible with.",
        always_export=True,
    )

    # TODO:
    # Perhaps we should version the plugin interface (not so the manifest, but
    # the actual mechanism/consumption of plugin information) independently
    # of napari itself

    on_activate: Optional[PythonName] = Field(
        default=None,
        description="Fully qualified python path to a function that will be called "
        "upon plugin activation (e.g. `'my_plugin.some_module:activate'`). The "
        "activate function can be used to connect command ids to python callables, or"
        " perform other side-effects. A plugin will be 'activated' when one of its "
        "contributions is requested by the user (such as a widget, or reader).",
    )
    _validate_activate_func = validator("on_activate", allow_reuse=True)(
        _validators.python_name
    )
    on_deactivate: Optional[PythonName] = Field(
        default=None,
        description="Fully qualified python path to a function that will be called "
        "when a user deactivates a plugin (e.g. `'my_plugin.some_module:deactivate'`)"
        ". This is optional, and may be used to perform any plugin cleanup.",
    )
    _validate_deactivate_func = validator("on_deactivate", allow_reuse=True)(
        _validators.python_name
    )

    contributions: ContributionPoints = Field(
        default_factory=ContributionPoints,
        description="An object describing the plugin's "
        "[contributions](./contributions)",
    )

    package_metadata: Optional[PackageMetadata] = Field(
        None,
        description="Package metadata following "
        "https://packaging.python.org/specifications/core-metadata/. "
        "For normal (non-dynamic) plugins, this data will come from the package's "
        "setup.cfg",
        hide_docs=True,
    )

    npe1_shim: bool = Field(
        False,
        description="Whether this manifest was created as a shim for an npe1 plugin.",
        hide_docs=True,
    )

    def __init__(self, **data):
        super().__init__(**data)
        if self.package_metadata is None and self.name:
            with suppress(metadata.PackageNotFoundError):
                meta = metadata.distribution(self.name).metadata
                self.package_metadata = PackageMetadata.from_dist_metadata(meta)

        if not self.npe1_shim:
            # assign plugin name on all contributions that have a private
            # _plugin_name field.
            for _, value in self.contributions or ():
                for item in value if isinstance(value, list) else [value]:
                    if isinstance(item, Executable):
                        item._plugin_name = self.name

    def __hash__(self):
        return hash((self.name, self.package_version))

    @property
    def license(self) -> Optional[str]:
        return self.package_metadata.license if self.package_metadata else None

    @property
    def package_version(self) -> Optional[str]:
        return self.package_metadata.version if self.package_metadata else None

    @property
    def description(self) -> Optional[str]:
        return self.package_metadata.summary if self.package_metadata else None

    @property
    def author(self) -> Optional[str]:
        return self.package_metadata.author if self.package_metadata else None

    @property
    def is_visible(self) -> bool:
        return self.visibility == "public"

    @validator("contributions", pre=True)
    def _coerce_none_contributions(cls, value):
        return [] if value is None else value

    @root_validator
    def _validate_root(cls, values: dict) -> dict:
        mf_name = values.get("name")

        # validate schema version
        declared_version = Version.parse(values.get("schema_version", ""))
        current_version = Version.parse(SCHEMA_VERSION)
        if current_version < declared_version:
            import warnings

            warnings.warn(
                f"The declared schema version for plugin {mf_name!r} "
                f"({declared_version}) is newer than npe2's schema version "
                f"({current_version}). Things may break, you should upgrade npe2.",
                stacklevel=2,
            )

        invalid_commands: List[str] = []
        if values.get("contributions") is not None:
            invalid_commands.extend(
                command.id
                for command in values["contributions"].commands or []
                if not command.id.startswith(f"{mf_name}.")
            )

        if invalid_commands:
            raise ValueError(
                "Commands identifiers must start with the current package name "
                f"followed by a dot: '{mf_name}'. The following commands do not: "
                f"{invalid_commands}"
            )

        if not values.get("display_name"):
            values["display_name"] = mf_name

        return values

    @classmethod
    def from_distribution(cls, name: str) -> PluginManifest:
        """Return PluginManifest given a distribution (package) name.

        Parameters
        ----------
        name : str
            Name of a python distribution installed in the environment.
            Note: this is the package name, not the top level module name,
            (e.g. "scikit-image", not "skimage").

        Returns
        -------
        PluginManifest
            The parsed manifest.

        Raises
        ------
        ValueError
            If the distribution exists, but does not provide a manifest
        PackageNotFoundError
            If there is no distribution found for `name`
        ValidationError
            If the manifest is not valid
        """
        dist = metadata.distribution(name)  # may raise PackageNotFoundError
        pm = _from_dist(dist)
        if not pm:
            raise ValueError(
                "Distribution {name!r} exists but does not provide a napari manifest"
            )
        return pm

    @classmethod
    def discover(
        cls, paths: Sequence[Union[str, Path]] = ()
    ) -> Iterator[DiscoverResults]:
        """Discover manifests in the environment.

        This function searches for installed python packages with a matching
        entry point group and then attempts to resolve the manifest file.

        The manifest file should be specified in the plugin's ``setup.cfg`` or
        ``setup.py`` file using the [entry point group][1]: "napari.manifest".
        For example, this would be the section for a plugin "npe-tester" with
        "napari.yaml" as the manifest file:

        .. code-block:: cfg

            [options.entry_points]
            napari.manifest =
                npe2-tester = npe2_tester:napari.yaml

        The manifest file is specified relative to the submodule root path.
        So for the example it will be loaded from:
        ``<path/to/npe2-tester>/napari.yaml``.

        [1]: https://packaging.python.org/specifications/entry-points/

        Parameters
        ----------
        paths : Sequence[str], optional
            paths to add to sys.path while discovering.

        Yields
        ------
        DiscoverResults: (3 namedtuples: manifest, entrypoint, error)
            3-tuples with either manifest or (entrypoint and error) being None.
        """
        with _temporary_path_additions(paths):
            for dist in metadata.distributions():
                try:
                    pm = _from_dist(dist)
                    if pm:
                        yield DiscoverResults(pm, dist, None)
                except ValidationError as e:
                    logger.warning(
                        "Invalid schema for package %r, please run"
                        " 'npe2 validate %s' to check for manifest errors.",
                        dist.metadata["Name"],
                        dist.metadata["Name"],
                    )
                    yield DiscoverResults(None, dist, e)

                except Exception as e:
                    logger.error(
                        "%s -> %r could not be imported: %s"
                        % (ENTRY_POINT, dist.metadata["Name"], e)
                    )
                    yield DiscoverResults(None, dist, e)

    @classmethod
    def _from_entrypoint(
        cls,
        entry_point: metadata.EntryPoint,
        distribution: Optional[metadata.Distribution] = None,
    ) -> PluginManifest:
        assert (match := entry_point.pattern.match(entry_point.value))
        module = match.group("module")

        spec = util.find_spec(module or "")
        if not spec:  # pragma: no cover
            raise ValueError(
                f"Cannot find module {module!r} declared in "
                f"entrypoint: {entry_point.value!r}"
            )

        assert (match := entry_point.pattern.match(entry_point.value))
        fname = match.group("attr")

        for loc in spec.submodule_search_locations or []:
            mf_file = Path(loc) / fname
            if mf_file.exists():
                mf = PluginManifest.from_file(mf_file)
                if distribution is not None:
                    meta = PackageMetadata.from_dist_metadata(distribution.metadata)
                    mf.package_metadata = meta

                    if mf.name != meta.name:
                        raise ValueError(  # pragma: no cover
                            f"The name field in the manifest ({mf.name!r}) "
                            f"must match the package name ({meta.name!r})"
                        )
                    return mf

        raise FileNotFoundError(  # pragma: no cover
            f"Could not find file {fname!r} in module {module!r}"
        )

    @classmethod
    def _from_package_or_name(
        cls, package_or_filename: Union[Path, str]
    ) -> PluginManifest:
        """Internal convenience function, calls both `from_file` and `from_distribution`

        Parameters
        ----------
        package_or_filename : Union[Path, str]
            Either a filename or a package name.  Will be tried first as a filename, and
            then as a distribution name.

        Returns
        -------
        PluginManifest
            [description]

        Raises
        ------
        ValidationError
            If the name can be resolved as either a distribution name or a file,
            but the manifest is not valid.
        ValueError
            If the name does not resolve to either a distribution name or a filename.

        """
        from pydantic import ValidationError

        from npe2 import PluginManifest

        try:
            return PluginManifest.from_file(package_or_filename)
        except ValidationError:  # pragma: no cover
            raise
        except (FileNotFoundError, ValueError):
            try:
                return PluginManifest.from_distribution(str(package_or_filename))
            except ValidationError:  # pragma: no cover
                raise
            except Exception as e:
                raise ValueError(
                    f"Could not find manifest for {package_or_filename!r} as either a "
                    "package name or a file.."
                ) from e

    def _serialized_data(self, **kwargs):
        kwargs.setdefault("exclude", {"package_metadata"})
        return super()._serialized_data(**kwargs)

    def validate_imports(self) -> None:
        """Checks recursively that all `python_name` fields are actually importable."""
        from .utils import import_python_name

        errors = []

        def check_pynames(m: BaseModel, loc=()):
            for name, value in m:
                if not value:
                    continue
                if isinstance(value, BaseModel):
                    return check_pynames(value, (*loc, name))
                field = m.__fields__[name]
                if isinstance(value, list) and isinstance(field.type_, ModelMetaclass):
                    return [check_pynames(i, (*loc, n)) for n, i in enumerate(value)]
                if field.outer_type_ is PythonName:
                    try:
                        import_python_name(value)
                    except (ImportError, AttributeError) as e:
                        errors.append(ErrorWrapper(e, (*loc, name)))

        check_pynames(self)
        if errors:
            raise ValidationError(errors, type(self))

    ValidationError = ValidationError  # for convenience of access


def _noop(*_, **__):
    return []  # pragma: no cover


@contextmanager
def discovery_blocked():
    orig = PluginManifest.discover
    PluginManifest.discover = _noop  # type: ignore [method-assign]
    try:
        yield
    finally:
        PluginManifest.discover = orig  # type: ignore [method-assign]


@contextmanager
def _temporary_path_additions(paths: Sequence[Union[str, Path]] = ()):
    if paths and (not isinstance(paths, Sequence) or isinstance(paths, str)):
        raise TypeError("paths must be a sequence of strings")  # pragma: no cover
    for p in reversed(paths):
        sys.path.insert(0, str(p))
    try:
        yield
    finally:
        for p in paths:
            sys.path.remove(str(p))


def _from_dist(dist: metadata.Distribution) -> Optional[PluginManifest]:
    """Return PluginManifest or NPE1Adapter for a metadata.Distribution object.

    ...depending on which entry points are available.
    """
    _npe1, _npe2 = [], None
    for ep in dist.entry_points:
        if ep.group == NPE1_ENTRY_POINT:
            _npe1.append(ep)
        elif ep.group == ENTRY_POINT:
            _npe2 = ep
    if _npe2:
        return PluginManifest._from_entrypoint(_npe2, dist)
    elif _npe1:
        from ._npe1_adapter import NPE1Adapter

        return NPE1Adapter(dist=dist)
    return None


if __name__ == "__main__":
    print(PluginManifest.schema_json(indent=2))



import pytest

from npe2 import DynamicPlugin, PluginManager
from npe2.manifest.contributions import SampleDataGenerator

TMP = "tmp"


@pytest.fixture
def tmp_plugin():
    local_pm = PluginManager()
    with DynamicPlugin(TMP, plugin_manager=local_pm) as tp:
        assert TMP in local_pm  # make sure it registered
        yield tp
    assert TMP not in local_pm  # make sure it cleaned up


def test_temporary_plugin(tmp_plugin: DynamicPlugin):
    """Test that we can use tmp_plugin to register commands for testing"""
    # everything is empty to begin with
    pm = tmp_plugin.plugin_manager
    contribs = tmp_plugin.manifest.contributions
    # everything is empty to begin with
    assert not contribs.commands
    assert not contribs.sample_data
    assert not contribs.readers
    assert not contribs.writers

    # we can populate with the contribute.x decorators

    @tmp_plugin.contribute.sample_data
    def make_image(x):
        return x

    @tmp_plugin.contribute.reader
    def read_path(path):
        ...

    # can override args
    ID = f"{TMP}.random_id"

    @tmp_plugin.contribute.command(id=ID)
    def some_command():
        return "hi!"

    # some require args

    with pytest.raises(AssertionError) as e:

        @tmp_plugin.contribute.writer
        def write_path_bad(path, layer_data):
            ...

    assert "layer_types must not be empty" in str(e.value)
    # it didn't get added
    assert "tmp.write_path_bad" not in pm.commands

    @tmp_plugin.contribute.writer(layer_types=["image"])
    def write_path(path, layer_data):
        ...

    # now it did
    assert "tmp.write_path" in pm.commands

    # contributions have been populated
    assert contribs.commands
    assert contribs.sample_data
    assert contribs.readers
    assert contribs.writers

    # and the commands work
    samples = next(contribs for plg, contribs in pm.iter_sample_data() if plg == TMP)
    gen = samples[0]
    assert isinstance(gen, SampleDataGenerator)
    assert gen.exec((1,), _registry=pm.commands) == 1

    cmd = pm.get_command(ID)
    assert cmd.exec(_registry=pm.commands) == "hi!"


def test_temporary_plugin_change_pm(tmp_plugin: DynamicPlugin):
    """We can change the plugin manager we're assigned to.

    Probably not necessary, but perhaps useful in tests.
    """
    start_pm = tmp_plugin.plugin_manager
    new_pm = PluginManager()

    @tmp_plugin.contribute.command
    def some_command():
        return "hi!"

    assert "tmp.some_command" in start_pm.commands
    assert "tmp.some_command" not in new_pm.commands

    tmp_plugin.plugin_manager = new_pm

    assert "tmp.some_command" not in start_pm.commands
    assert "tmp.some_command" in new_pm.commands

    tmp_plugin.clear()
    assert not tmp_plugin.manifest.contributions.commands


def test_temporary_plugin_spawn(tmp_plugin: DynamicPlugin):
    new = tmp_plugin.spawn("another-name", register=True)
    assert new.name == "another-name"
    assert new.display_name == "another-name"
    assert new.plugin_manager == tmp_plugin.plugin_manager

    assert (t1 := tmp_plugin.spawn(register=True)).name == f"{tmp_plugin.name}-1"
    assert (t2 := tmp_plugin.spawn()).name == f"{tmp_plugin.name}-2"

    assert t1.name in tmp_plugin.plugin_manager._manifests
    assert t2.name not in tmp_plugin.plugin_manager._manifests



from __future__ import annotations

from dataclasses import dataclass
from functools import partial
from typing import TYPE_CHECKING, Any, Callable, Dict, Optional, Union

from psygnal import Signal

from .manifest import utils
from .manifest._validators import DOTTED_NAME_PATTERN
from .types import PythonName

PDisposable = Callable[[], None]

if TYPE_CHECKING:
    from .manifest.schema import PluginManifest


@dataclass
class CommandHandler:
    id: str
    function: Optional[Callable] = None
    python_name: Optional[PythonName] = None

    def resolve(self) -> Callable:
        if self.function is not None:
            return self.function
        if self.python_name is None:
            raise RuntimeError("cannot resolve command without python_name")

        try:
            self.function = utils.import_python_name(self.python_name)
        except Exception as e:
            raise RuntimeError(
                f"Failed to import command at {self.python_name!r}: {e}"
            ) from e

        return self.function


class CommandRegistry:
    command_registered = Signal(str)
    command_unregistered = Signal(str)

    def __init__(self) -> None:
        self._commands: Dict[str, CommandHandler] = {}

    def register(self, id: str, command: Union[Callable, str]) -> PDisposable:
        """Register a command under `id`.

        Parameters
        ----------
        id : str
            A unique key with which to refer to this command
        command : Union[Callable, str]
            Either a callable object, or (if a string) the fully qualified name of a
            python object.  If a string is provided, it is not imported until
            the command is actually executed.

        Returns
        -------
        PDisposable
            A callable that, when called, unregisters the command.

        Raises
        ------
        ValueError
            If the id is not a non-empty string, or if it already exists.
        TypeError
            If `command` is not a string or a callable object.
        """
        if not (isinstance(id, str) and id.strip()):
            raise ValueError(
                f"Invalid command id for {command}, must be non-empty string"
            )
        if id in self._commands:
            raise ValueError(f"Command {id} already exists")

        if isinstance(command, str):
            if not DOTTED_NAME_PATTERN.match(command):
                raise ValueError(
                    "String command {command!r} is not a valid qualified python path."
                )
            cmd = CommandHandler(id, python_name=PythonName(command))
        elif not callable(command):
            raise TypeError(f"Cannot register non-callable command: {command}")
        else:
            cmd = CommandHandler(id, function=command)

        # TODO: validate arguments and type constraints
        # possibly wrap command in a type validator?

        self._commands[id] = cmd
        self.command_registered.emit(id)

        return partial(self.unregister, id)

    def unregister(self, id: str):
        """Unregister command with key `id`.  No-op if key doesn't exist."""
        if id in self._commands:
            del self._commands[id]
            self.command_unregistered.emit(id)

    def register_manifest(self, mf: PluginManifest) -> None:
        """Register all commands in a manifest"""
        if mf.contributions and mf.contributions.commands:
            for cmd in mf.contributions.commands:
                if cmd.python_name and cmd.id not in self:
                    self.register(cmd.id, cmd.python_name)

    def unregister_manifest(self, mf: PluginManifest) -> None:
        """Unregister all commands in a manifest"""
        if mf.contributions and mf.contributions.commands:
            for cmd in mf.contributions.commands:
                if cmd.id in self:
                    self.unregister(cmd.id)

    def get(self, id: str) -> Callable:
        """Get callable object for command `id`."""
        # FIXME: who should control activation?
        if id not in self._commands:
            from ._plugin_manager import PluginManager

            pm = PluginManager.instance()

            if id in pm._contrib._commands:
                _, plugin_key = pm._contrib._commands[id]
                pm.activate(plugin_key)
            if id not in self._commands:  # sourcery skip
                raise KeyError(f"command {id!r} not registered")
        return self._commands[id].resolve()

    def execute(self, id: str, args=(), kwargs=None) -> Any:
        if kwargs is None:
            kwargs = {}
        return self.get(id)(*args, **kwargs)

    def __contains__(self, id: str):
        return id in self._commands



from ._compile import compile
from ._visitors import (
    NPE2PluginModuleVisitor,
    find_npe1_module_contributions,
    find_npe2_module_contributions,
)

__all__ = [
    "NPE2PluginModuleVisitor",
    "find_npe2_module_contributions",
    "find_npe1_module_contributions",
    "compile",
]



from ._commands import CommandContribution
from ._configuration import ConfigurationContribution, ConfigurationProperty
from ._contributions import ContributionPoints
from ._menus import MenuCommand, MenuItem, Submenu
from ._readers import ReaderContribution
from ._sample_data import SampleDataContribution, SampleDataGenerator, SampleDataURI
from ._submenu import SubmenuContribution
from ._themes import ThemeColors, ThemeContribution
from ._widgets import WidgetContribution
from ._writers import LayerType, LayerTypeConstraint, WriterContribution

__all__ = [
    "CommandContribution",
    "ConfigurationContribution",
    "ConfigurationProperty",
    "ContributionPoints",
    "LayerType",
    "LayerTypeConstraint",
    "MenuCommand",
    "MenuItem",
    "ReaderContribution",
    "SampleDataContribution",
    "SampleDataGenerator",
    "SampleDataURI",
    "Submenu",
    "SubmenuContribution",
    "ThemeColors",
    "ThemeContribution",
    "WidgetContribution",
    "WriterContribution",
]



import pytest

from npe2.manifest.contributions import ContributionPoints
from npe2.manifest.schema import PluginManifest
from npe2.manifest.utils import (
    Version,
    deep_update,
    merge_contributions,
    merge_manifests,
)


def test_version():
    v = Version.parse(b"0.1.2")

    assert v == "0.1.2"
    assert v > {"major": 0, "minor": 1, "patch": 0}
    assert v <= (0, 2, 0)
    assert v == Version(0, 1, 2)
    assert list(v) == [0, 1, 2, None, None]
    assert str(v) == "0.1.2"

    with pytest.raises(TypeError):
        assert v == 1.2

    with pytest.raises(ValueError):
        Version.parse("alkfdjs")

    with pytest.raises(TypeError):
        Version.parse(1.2)  # type: ignore


def test_merge_manifests():
    with pytest.raises(ValueError):
        merge_manifests([])

    with pytest.raises(AssertionError) as e:
        merge_manifests([PluginManifest(name="p1"), PluginManifest(name="p2")])
    assert "All manifests must have same name" in str(e.value)

    pm1 = PluginManifest(
        name="plugin",
        contributions={
            "commands": [{"id": "plugin.command", "title": "some writer"}],
            "writers": [{"command": "plugin.command", "layer_types": ["image"]}],
        },
    )
    pm2 = PluginManifest(
        name="plugin",
        contributions={
            "commands": [{"id": "plugin.command", "title": "some reader"}],
            "readers": [{"command": "plugin.command", "filename_patterns": [".tif"]}],
        },
    )
    expected_merge = PluginManifest(
        name="plugin",
        contributions={
            "commands": [
                {"id": "plugin.command", "title": "some writer"},
                {"id": "plugin.command_2", "title": "some reader"},  # no dupes
            ],
            "writers": [{"command": "plugin.command", "layer_types": ["image"]}],
            "readers": [{"command": "plugin.command_2", "filename_patterns": [".tif"]}],
        },
    )

    assert merge_manifests([pm1]) is pm1
    assert merge_manifests([pm1, pm2]) == expected_merge


def test_merge_contributions():
    a = ContributionPoints(
        commands=[
            {"id": "plugin.command", "title": "some writer"},
        ],
        writers=[{"command": "plugin.command", "layer_types": ["image"]}],
    )
    b = ContributionPoints(
        commands=[
            {"id": "plugin.command", "title": "some writer"},
        ],
        writers=[{"command": "plugin.command", "layer_types": ["image"]}],
    )
    c = ContributionPoints(
        commands=[
            {"id": "plugin.command", "title": "some writer"},
        ],
        writers=[{"command": "plugin.command", "layer_types": ["image"]}],
    )
    expected = ContributionPoints(
        commands=[
            {"id": "plugin.command", "title": "some writer"},
            {"id": "plugin.command_2", "title": "some writer"},
            {"id": "plugin.command_3", "title": "some writer"},
        ],
        writers=[
            {"command": "plugin.command", "layer_types": ["image"]},
            {"command": "plugin.command_2", "layer_types": ["image"]},
            {"command": "plugin.command_3", "layer_types": ["image"]},
        ],
    )

    d = ContributionPoints(**merge_contributions((a, b, c)))
    assert d == expected

    # with overwrite, later contributions with matching command ids take precendence.
    e = ContributionPoints(**merge_contributions((a, b, c), overwrite=True))
    expected = ContributionPoints(
        commands=[
            {"id": "plugin.command", "title": "some writer"},
        ],
        writers=[
            {"command": "plugin.command", "layer_types": ["image"]},
        ],
    )
    assert e == a


def test_deep_update():
    a = {"a": {"b": 1, "c": 2}, "e": 2}
    b = {"a": {"d": 4, "c": 3}, "f": 0}
    c = deep_update(a, b, copy=True)
    assert c == {"a": {"b": 1, "d": 4, "c": 3}, "e": 2, "f": 0}
    assert a == {"a": {"b": 1, "c": 2}, "e": 2}

    deep_update(a, b, copy=False)
    assert a == {"a": {"b": 1, "d": 4, "c": 3}, "e": 2, "f": 0}



from typing import (
    TYPE_CHECKING,
    Callable,
    Dict,
    List,
    Literal,
    NewType,
    Optional,
    Protocol,
    Sequence,
    Tuple,
    Union,
)

if TYPE_CHECKING:
    import magicgui.widgets
    import numpy as np
    import qtpy.QtWidgets


# General types

# PathLike = Union[str, pathlib.Path]  # we really have to pick one
PathLike = str
PathOrPaths = Union[PathLike, Sequence[PathLike]]
PythonName = NewType("PythonName", str)

# Layer-related types


class ArrayLike(Protocol):
    shape: Tuple[int, ...]
    ndim: int
    dtype: "np.dtype"

    def __array__(self) -> "np.ndarray":
        ...  # pragma: no cover


LayerName = Literal[
    "image", "labels", "points", "shapes", "surface", "tracks", "vectors"
]
Metadata = Dict
DataType = Union[ArrayLike, Sequence[ArrayLike]]
FullLayerData = Tuple[DataType, Metadata, LayerName]
LayerData = Union[Tuple[DataType], Tuple[DataType, Metadata], FullLayerData]

# ########################## CONTRIBUTIONS #################################

# WidgetContribution.command must point to a WidgetCreator
Widget = Union["magicgui.widgets.Widget", "qtpy.QtWidgets.QWidget"]
WidgetCreator = Callable[..., Widget]

# ReaderContribution.command must point to a ReaderGetter
ReaderFunction = Callable[[PathOrPaths], List[LayerData]]
ReaderGetter = Callable[[PathOrPaths], Optional[ReaderFunction]]


# SampleDataGenerator.command must point to a SampleDataCreator
SampleDataCreator = Callable[..., List[LayerData]]

# WriterContribution.command must point to a WriterFunction
# Writers that take at most one layer must provide a SingleWriterFunction command.
# Otherwise, they must provide a MultiWriterFunction.
# where the number of layers they take is defined as
# n = sum(ltc.max() for ltc in WriterContribution.layer_type_constraints())
SingleWriterFunction = Callable[[str, DataType, Metadata], List[str]]
MultiWriterFunction = Callable[[str, List[FullLayerData]], List[str]]
WriterFunction = Union[SingleWriterFunction, MultiWriterFunction]

# ##########################################################################



import os
from importlib import metadata
from subprocess import CalledProcessError
from typing import TYPE_CHECKING

import pytest

from npe2._inspection._full_install import isolated_plugin_env
from npe2.cli import app

if TYPE_CHECKING:
    from pathlib import Path

PLUGIN: str = os.getenv("TEST_PACKAGE_NAME") or ""
if not PLUGIN:
    pytest.skip("skipping plugin specific tests", allow_module_level=True)


@pytest.fixture(scope="session")
def plugin_env():
    try:
        with isolated_plugin_env(PLUGIN) as env:
            yield env
    except CalledProcessError as e:
        if "Failed building wheel" in str(e.output):
            yield None


def test_entry_points_importable(plugin_env):
    if plugin_env is None:
        pytest.mark.xfail()
        return

    entry_points = [
        ep
        for ep in metadata.distribution(PLUGIN).entry_points
        if ep.group in ("napari.plugin", "napari.manifest")
    ]
    if PLUGIN not in {"napari-console", "napari-error-reporter"}:
        assert entry_points

    for ep in entry_points:
        if ep.group == "napari.plugin":
            ep.load()


def test_fetch(tmp_path: "Path"):
    from typer.testing import CliRunner

    mf_file = tmp_path / "manifest.yaml"

    result = CliRunner().invoke(app, ["fetch", PLUGIN, "-o", str(mf_file)])
    assert result.exit_code == 0
    assert PLUGIN in mf_file.read_text()
    result2 = CliRunner().invoke(app, ["validate", str(mf_file)])
    assert result2.exit_code == 0



import contextlib
from inspect import Parameter, Signature
from typing import Any, Callable, List, Sequence, Type, TypeVar

from pydantic import BaseModel

from .manifest import contributions

__all__ = [
    "CHECK_ARGS_PARAM",
    "on_activate",
    "on_deactivate",
    "reader",
    "sample_data_generator",
    "widget",
    "writer",
]


T = TypeVar("T", bound=Callable[..., Any])

CHECK_ARGS_PARAM = "ensure_args_valid"


def _build_decorator(contrib: Type[BaseModel]) -> Callable:
    """Create a decorator (e.g. `@implements.reader`) to mark an object as a contrib.

    Parameters
    ----------
    contrib : Type[BaseModel]
        The type of contribution this object implements.
    """
    # build a signature based on the fields in this contribution type, mixed with
    # the fields in the CommandContribution
    contribs: Sequence[Type[BaseModel]] = (contributions.CommandContribution, contrib)
    params: List[Parameter] = []
    for contrib in contribs:
        # iterate over the fields in the contribution types
        for field in contrib.__fields__.values():
            # we don't need python_name (since that will be gleaned from the function
            # we're decorating) ... and we don't need `command`, since that will just
            # be a string pointing to the contributions.commands entry that we are
            # creating here.
            if field.name not in {"python_name", "command"}:
                # ensure that required fields raise a TypeError if they are not provided
                default = Parameter.empty if field.required else field.get_default()
                # create the parameter and add it to the signature.
                param = Parameter(
                    field.name,
                    Parameter.KEYWORD_ONLY,
                    default=default,
                    annotation=field.outer_type_ or field.type_,
                )
                params.append(param)

    # add one more parameter to control whether the arguments in the decorator itself
    # are validated at runtime
    params.append(
        Parameter(
            CHECK_ARGS_PARAM,
            kind=Parameter.KEYWORD_ONLY,
            default=False,
            annotation=bool,
        )
    )

    signature = Signature(parameters=params, return_annotation=Callable[[T], T])

    # creates the actual `@npe2.implements.something` decorator
    # this just stores the parameters for the corresponding contribution type
    # as attributes on the function being decorated.
    def _deco(**kwargs) -> Callable[[T], T]:
        def _store_attrs(func: T) -> T:
            # If requested, assert that we've satisfied the signature when
            # the decorator is invoked at runtime.
            # TODO: improve error message to provide context
            if kwargs.pop(CHECK_ARGS_PARAM, False):
                signature.bind(**kwargs)

            # TODO: check if it's already there and assert the same id
            # store these attributes on the function
            with contextlib.suppress(AttributeError):
                setattr(func, f"_npe2_{contrib.__name__}", kwargs)

            # return the original decorated function
            return func

        return _store_attrs

    # set the signature and return the decorator
    _deco.__signature__ = signature  # type: ignore
    return _deco


# builds decorators for each of the contribution types that are essentially just
# pointers to some command.
reader = _build_decorator(contributions.ReaderContribution)
writer = _build_decorator(contributions.WriterContribution)
widget = _build_decorator(contributions.WidgetContribution)
sample_data_generator = _build_decorator(contributions.SampleDataGenerator)


def on_activate(func):
    """Mark a function to be called when a plugin is activated."""
    func.npe2_on_activate = True
    return func


def on_deactivate(func):
    """Mark a function to be called when a plugin is deactivated."""
    func.npe2_on_deactivate = True
    return func



from __future__ import annotations

import contextlib
import os
import warnings
from collections import Counter
from fnmatch import fnmatch
from importlib import metadata
from pathlib import Path
from typing import (
    TYPE_CHECKING,
    AbstractSet,
    Any,
    Callable,
    DefaultDict,
    Dict,
    Iterable,
    Iterator,
    List,
    Mapping,
    Optional,
    Sequence,
    Set,
    Tuple,
    Union,
)

from psygnal import Signal, SignalGroup

from ._command_registry import CommandRegistry
from .manifest import PluginManifest
from .manifest._npe1_adapter import NPE1Adapter
from .manifest.contributions import LayerType, WriterContribution
from .types import PathLike, PythonName

if TYPE_CHECKING:
    from .manifest.contributions import (
        CommandContribution,
        MenuItem,
        ReaderContribution,
        SampleDataContribution,
        SubmenuContribution,
        ThemeContribution,
        WidgetContribution,
    )

    IntStr = Union[int, str]
    AbstractSetIntStr = AbstractSet[IntStr]
    DictIntStrAny = Dict[IntStr, Any]
    MappingIntStrAny = Mapping[IntStr, Any]
    InclusionSet = Union[AbstractSetIntStr, MappingIntStrAny, None]
    DisposeFunction = Callable[[], None]

__all__ = ["PluginContext", "PluginManager"]
PluginName = str  # this is `PluginManifest.name`


class _ContributionsIndex:
    def __init__(self) -> None:
        self._indexed: Set[str] = set()
        self._commands: Dict[str, Tuple[CommandContribution, PluginName]] = {}
        self._readers: List[Tuple[str, ReaderContribution]] = []
        self._writers: List[Tuple[LayerType, int, int, WriterContribution]] = []

        # DEPRECATED: only here for napari <= 0.4.15 compat.
        self._samples: DefaultDict[str, List[SampleDataContribution]] = DefaultDict(
            list
        )

    def reindex(self, manifest):
        self.remove_contributions(manifest.name)
        self.index_contributions(manifest)

    def index_contributions(self, manifest: PluginManifest):
        ctrb = manifest.contributions
        if not ctrb or manifest.name in self._indexed:
            return  # pragma: no cover

        self._indexed.add(manifest.name)
        for cmd in ctrb.commands or ():
            self._commands[cmd.id] = cmd, manifest.name
        for reader in ctrb.readers or ():
            for pattern in reader.filename_patterns:
                self._readers.append((pattern, reader))
            if reader.accepts_directories:
                self._readers.append(("", reader))
        for writer in ctrb.writers or ():
            for c in writer.layer_type_constraints():
                self._writers.append((c.layer_type, *c.bounds, writer))

        # DEPRECATED: only here for napari <= 0.4.15 compat.
        if ctrb.sample_data:
            self._samples[manifest.name] = ctrb.sample_data

    def remove_contributions(self, key: PluginName) -> None:
        """This must completely remove everything added by `index_contributions`."""
        if key not in self._indexed:
            return  # pragma: no cover

        for cmd_id, (_, plugin) in list(self._commands.items()):
            if key == plugin:
                del self._commands[cmd_id]

        self._readers = [
            (pattern, reader)
            for pattern, reader in self._readers
            if reader.plugin_name != key
        ]

        self._writers = [
            (layer_type, min_, max_, writer)
            for layer_type, min_, max_, writer in self._writers
            if writer.plugin_name != key
        ]

        self._indexed.remove(key)

        # DEPRECATED: only here for napari <= 0.4.15 compat.
        self._samples.pop(key, None)

    def get_command(self, command_id: str) -> CommandContribution:
        return self._commands[command_id][0]

    def iter_compatible_readers(self, paths: List[str]) -> Iterator[ReaderContribution]:
        assert isinstance(paths, list)
        if not paths:
            return  # pragma: no cover

        if len({Path(i).suffix for i in paths}) > 1:
            raise ValueError(
                "All paths in the stack list must have the same extension."
            )
        path = paths[0]
        if not path:
            return
        assert isinstance(path, str)

        if os.path.isdir(path):
            yield from (r for pattern, r in self._readers if pattern == "")
        else:
            # not sure about the set logic as it won't be lazy anymore,
            # but would we yield duplicate anymore.
            # above does not have have the unseen check either.
            # it's easy to make an iterable version if we wish, or use more-itertools.
            yield from {r for pattern, r in self._readers if fnmatch(path, pattern)}

    def iter_compatible_writers(
        self, layer_types: Sequence[str]
    ) -> Iterator[WriterContribution]:
        """Attempt to match writers that consume all layers."""

        if not layer_types:
            return

        # First count how many of each distinct type are requested. We'll use
        # this to get candidate writers compatible with the requested count.
        counts = Counter(layer_types)

        def _get_candidates(lt: LayerType) -> Set[WriterContribution]:
            return {
                w
                for layer, min_, max_, w in self._writers
                if layer == lt and (min_ <= counts[lt] < max_)
            }

        # keep ordered without duplicates
        candidates = list({w: None for _, _, _, w in self._writers})
        for lt in LayerType:
            if candidates:
                candidates = [i for i in candidates if i in _get_candidates(lt)]
            else:
                break

        def _writer_key(writer: WriterContribution) -> Tuple[bool, int]:
            # 1. writers with no file extensions (like directory writers) go last
            no_ext = len(writer.filename_extensions) == 0

            # 2. more "specific" writers first
            nbounds = sum(not c.is_zero() for c in writer.layer_type_constraints())
            return (no_ext, nbounds)

        yield from sorted(candidates, key=_writer_key)


class PluginManagerEvents(SignalGroup):
    plugins_registered = Signal(
        set,
        description="Emitted with a set of PluginManifest instances "
        "whenever new plugins are registered. 'Registered' means that a "
        "manifest has been provided or discovered.",
    )
    activation_changed = Signal(
        set,
        set,
        description="Emitted with two arguments: a set of plugin "
        "names that were activated, and a set of names that were "
        "deactivated. 'Activated' means the plugin has been *imported*, its "
        "`on_activate` function was called.",
    )
    enablement_changed = Signal(
        set,
        set,
        description="Emitted with two arguments: a set of plugin names "
        "that were enabled, and a set of names that were "
        "disabled. 'Disabled' means the plugin remains installed, but it "
        "cannot be activated, and its contributions will not be indexed.",
    )


class PluginManager:
    __instance: Optional[PluginManager] = None  # a global instance
    _contrib: _ContributionsIndex
    events: PluginManagerEvents

    def __init__(
        self, *, disable: Iterable[str] = (), reg: Optional[CommandRegistry] = None
    ) -> None:
        self._disabled_plugins: Set[PluginName] = set(disable)
        self._command_registry = reg or CommandRegistry()
        self._contexts: Dict[PluginName, PluginContext] = {}
        self._contrib = _ContributionsIndex()
        self._manifests: Dict[PluginName, PluginManifest] = {}
        self.events = PluginManagerEvents(self)
        self._npe1_adapters: List[NPE1Adapter] = []

        # up to napari 0.4.15, discovery happened in the init here
        # so if we're running on an older version of napari, we need to discover
        try:
            nv = metadata.version("napari")
        except metadata.PackageNotFoundError:  # pragma: no cover
            pass
        else:  # pragma: no cover
            vsplit = nv.split(".")[:4]
            if (
                "dev" in nv
                and vsplit < ["0", "4", "16", "dev4"]
                or "dev" not in nv
                and vsplit < ["0", "4", "16"]
            ):
                self.discover()

    @classmethod
    def instance(cls) -> PluginManager:
        """Return global PluginManager singleton instance."""
        if cls.__instance is None:
            cls.__instance = cls()
        return cls.__instance

    @property
    def commands(self) -> CommandRegistry:
        return self._command_registry

    # Discovery, activation, enablement

    def discover(
        self, paths: Sequence[str] = (), clear=False, include_npe1=False
    ) -> int:
        """Discover and index plugin manifests in the environment.

        Parameters
        ----------
        paths : Sequence[str]
            Optional list of strings to insert at front of sys.path when
            discovering.
        clear : bool
            Clear and re-index the environment.  If `False` (the default),
            calling discover again will only register and index newly
            discovered plugins. (Existing manifests will not be re-indexed)
        include_npe1 : bool
            Whether to detect npe1 plugins as npe1_adapters during discovery.
            By default `False`.

        Returns
        -------
        discover_count : int
            Number of discovered plugins

        """
        if clear:
            self._contrib = _ContributionsIndex()
            self._manifests.clear()

        count = 0

        with self.events.plugins_registered.paused(lambda a, b: (a[0] | b[0],)):
            for result in PluginManifest.discover(paths=paths):
                if (
                    result.manifest
                    and result.manifest.name not in self._manifests
                    and (include_npe1 or not isinstance(result.manifest, NPE1Adapter))
                ):
                    self.register(result.manifest, warn_disabled=False)
                    count += 1
        return count

    def index_npe1_adapters(self):
        """Import and index any/all npe1 adapters."""
        with warnings.catch_warnings():
            warnings.showwarning = lambda e, *_: print(str(e).split(" Please add")[0])
            while self._npe1_adapters:
                self._contrib.index_contributions(self._npe1_adapters.pop())

    def register(
        self, manifest_or_package: Union[PluginManifest, str], warn_disabled=True
    ) -> None:
        """Register a plugin manifest, path to manifest file, or a package name.

        Parameters
        ----------
        manifest_or_package : Union[PluginManifest, str]
            Either a PluginManifest instance or a string. If a string, should be either
            the name of a plugin package, or a path to a plugin manifest file.
        warn_disabled : bool, optional
            If True, emits a warning if the plugin being registered is marked as
            disabled, by default True.

        Raises
        ------
        ValueError
            If a plugin with the same name is already registered.
        """
        if isinstance(manifest_or_package, str):
            if Path(manifest_or_package).is_file():
                manifest = PluginManifest.from_file(manifest_or_package)
            else:
                manifest = PluginManifest.from_distribution(manifest_or_package)
        elif isinstance(manifest_or_package, PluginManifest):
            manifest = manifest_or_package
        else:  # pragma: no cover
            raise TypeError(
                "The first argument to register must be a string or a PluginManifest."
            )
        if manifest.name in self._manifests:
            raise ValueError(f"A manifest with name {manifest.name!r} already exists.")

        self._manifests[manifest.name] = manifest
        if self.is_disabled(manifest.name):
            if warn_disabled:
                warnings.warn(
                    f"Disabled plugin {manifest.name!r} was registered, but will not "
                    "be indexed. Use `warn_disabled=False` to suppress this message.",
                    stacklevel=2,
                )
        elif isinstance(manifest, NPE1Adapter):
            self._npe1_adapters.append(manifest)
        else:
            self._contrib.index_contributions(manifest)
        self.events.plugins_registered.emit({manifest})

    def unregister(self, key: PluginName):
        """Unregister plugin named `key`."""
        if key not in self._manifests:
            raise ValueError(f"No registered plugin named {key!r}")  # pragma: no cover
        self.deactivate(key)
        self._contrib.remove_contributions(key)
        self._manifests.pop(key)

    def activate(self, key: PluginName) -> PluginContext:
        """Activate plugin with `key`.

        This does the following:
            - finds the manifest for the associated plugin key
            - gets or creates a PluginContext for the plugin
            - bails if it's already activated
            - otherwise calls the plugin's activate() function, passing the Context.
            - imports any commands that were declared as python_name:
            - emits an event
        """
        # TODO: this is an important function... should be carefully considered
        if key not in self._manifests:
            raise KeyError(f"Cannot activate unrecognized plugin: {key!r}")

        if self.is_disabled(key):
            raise ValueError(f"Cannot activate disabled plugin: {key!r}")

        # create the context that will be with this plugin for its lifetime.
        ctx = self.get_context(key)
        if ctx._activated:
            # prevent "reactivation"
            return ctx

        mf = self._manifests[key]
        try:
            if mf.on_activate:
                _call_python_name(mf.on_activate, args=(ctx,))
        except Exception as e:  # pragma: no cover
            self._contexts.pop(key, None)
            raise type(e)(f"Activating plugin {key!r} failed: {e}") from e

        self.commands.register_manifest(mf)
        ctx._activated = True
        self.events.activation_changed({mf.name}, {})
        return ctx

    def get_context(self, plugin_name: PluginName) -> PluginContext:
        """Return PluginContext for plugin_name"""
        if plugin_name not in self._contexts:
            self._contexts[plugin_name] = PluginContext(plugin_name, reg=self.commands)
        return self._contexts[plugin_name]

    def deactivate(self, plugin_name: PluginName) -> None:
        """Deactivate `plugin_name`

        This does the following:
            - unregisters all commands from the associated manifest
            - calls the plugin's on_deactivate() func, passing the Context.
            - calls and cleanup functions in the context's `_dispose` method.
            - emits an event

        This does not:
            - "unindex" contributions (i.e. the contributions of a deactivated plugin
              are still visible in the index)
            - "disable" the plugin (i.e. it can still be used).
        """
        mf = self._manifests[plugin_name]
        self.commands.unregister_manifest(mf)
        if plugin_name not in self._contexts:
            return
        ctx = self._contexts.pop(plugin_name)
        if mf.on_deactivate:
            _call_python_name(mf.on_deactivate, args=(ctx,))
        ctx._activated = False
        ctx._dispose()
        self.events.activation_changed({}, {mf.name})

    def enable(self, plugin_name: PluginName) -> None:
        """Enable a plugin (which mostly means just `un-disable` it).

        This is a no-op if the plugin wasn't already disabled.
        """
        if not self.is_disabled(plugin_name):
            return  # pragma: no cover

        self._disabled_plugins.remove(plugin_name)
        mf = self._manifests.get(plugin_name)
        if mf is not None:
            self._contrib.index_contributions(mf)
        self.events.enablement_changed({plugin_name}, {})

    def disable(self, plugin_name: PluginName) -> None:
        """Disable a plugin.

        'Disabled' means the plugin remains installed, but it cannot be activated,
        and its contributions will not be indexed.  Menu items and keybindings and
        such will not be available.

        In napari, plugin disablement is persisted across sessions.
        """
        if self.is_disabled(plugin_name):
            return  # pragma: no cover

        with contextlib.suppress(KeyError):
            self.deactivate(plugin_name)

        self._disabled_plugins.add(plugin_name)
        self._contrib.remove_contributions(plugin_name)
        self.events.enablement_changed({}, {plugin_name})

    def is_disabled(self, plugin_name: str) -> bool:
        """Return `True` if plugin_name is disabled."""
        return plugin_name in self._disabled_plugins

    # Getting manifests

    def get_manifest(self, plugin_name: str) -> PluginManifest:
        """Get manifest for `plugin_name`"""
        key = str(plugin_name).split(".")[0]
        if key not in self._manifests:
            msg = f"Manifest key {key!r} not found in {list(self._manifests)}"
            raise KeyError(msg)
        return self._manifests[key]

    def iter_manifests(
        self, disabled: Optional[bool] = None
    ) -> Iterator[PluginManifest]:
        """Iterate through registered manifests.

        Parameters
        ----------
        disabled : Optional[bool]
            If `False`, yield only enabled manifests.  If `True`, yield only disabled
            manifests.  If `None` (the default), yield all manifests.

        Yields
        ------
        PluginManifest
        """
        for key, mf in self._manifests.items():
            if disabled is True and not self.is_disabled(key):
                continue
            elif disabled is False and self.is_disabled(key):
                continue
            yield mf

    def dict(
        self,
        *,
        include: Optional[InclusionSet] = None,
        exclude: Optional[InclusionSet] = None,
    ) -> Dict[str, Any]:
        """Return a dictionary with the state of the plugin manager.

        `include` and `exclude` will be passed to each `PluginManifest.dict()`
        See pydantic documentation for details:
        https://pydantic-docs.helpmanual.io/usage/exporting_models/#modeldict

        `include` and `exclude` may be a set of dotted strings, indicating
        nested fields in the manifest model.  For example:

            {'contributions.readers', 'package_metadata.description'}

        will be expanded to

            {
                'contributions': {'readers': True},
                'package_metadata': {'description': True}
            }

        This facilitates selection of nested fields on the command line.


        Parameters
        ----------
        include : InclusionSet, optional
            A set of manifest fields to include, by default all fields are included.
        exclude : InclusionSet, optional
            A set of manifest fields to exclude, by default no fields are excluded.

        Returns
        -------
        Dict[str, Any]
            Dictionary with the state of the plugin manager.  Keys will include

                - `'plugins'`: dict of `{name: manifest.dict()} for discovered plugins
                - `'disabled'`: set of disabled plugins
                - `'activated'`: set of activated plugins

        """
        # _include =
        out: Dict[str, Any] = {
            "plugins": {
                mf.name: mf.dict(
                    include=_expand_dotted_set(include),
                    exclude=_expand_dotted_set(exclude),
                )
                for mf in self.iter_manifests()
            }
        }
        if not exclude or "disabled" not in exclude:
            out["disabled"] = set(self._disabled_plugins)
        if not exclude or "activated" not in exclude:
            out["activated"] = {
                name for name, ctx in self._contexts.items() if ctx._activated
            }
        return out

    def __contains__(self, name: str) -> bool:
        return name in self._manifests

    def __getitem__(self, name: str) -> PluginManifest:
        return self.get_manifest(name)

    def __len__(self) -> int:
        return len(self._manifests)

    # Accessing Contributions

    def get_command(self, command_id: str) -> CommandContribution:
        """Retrieve CommandContribution for `command_id`"""
        return self._contrib.get_command(command_id)

    def get_submenu(self, submenu_id: str) -> SubmenuContribution:
        """Get SubmenuContribution for `submenu_id`."""
        for mf in self.iter_manifests(disabled=False):
            for subm in mf.contributions.submenus or ():
                if submenu_id == subm.id:
                    return subm
        raise KeyError(f"No plugin provides a submenu with id {submenu_id}")

    def iter_menu(self, menu_key: str, disabled=False) -> Iterator[MenuItem]:
        """Iterate over `MenuItems` in menu with id `menu_key`."""
        for mf in self.iter_manifests(disabled=disabled):
            yield from mf.contributions.menus.get(menu_key, ())

    def menus(self, disabled=False) -> Dict[str, List[MenuItem]]:
        """Return all registered menu_key -> List[MenuItems]."""
        _menus: DefaultDict[str, List[MenuItem]] = DefaultDict(list)
        for mf in self.iter_manifests(disabled=disabled):
            for key, menus in mf.contributions.menus.items():
                _menus[key].extend(menus)
        return dict(_menus)

    def iter_themes(self) -> Iterator[ThemeContribution]:
        """Iterate over discovered/enuabled `ThemeContributions`."""
        for mf in self.iter_manifests(disabled=False):
            yield from mf.contributions.themes or ()

    def iter_compatible_readers(
        self, path: Union[PathLike, Sequence[str]]
    ) -> Iterator[ReaderContribution]:
        """Iterate over ReaderContributions compatible with `path`.

        Parameters
        ----------
        path : Union[PathLike, Sequence[str]]
            Pathlike or list of pathlikes, with file(s) to read.
        """
        if isinstance(path, (str, Path)):
            path = [path]
        assert isinstance(path, list)
        return self._contrib.iter_compatible_readers(path)

    def iter_compatible_writers(
        self, layer_types: Sequence[str]
    ) -> Iterator[WriterContribution]:
        """Iterate over compatible WriterContributions given a sequence of layer_types.

        Parameters
        ----------
        layer_types : Sequence[str]
            list of lowercase Layer type names like `['image', 'labels']`
        """
        return self._contrib.iter_compatible_writers(layer_types)

    def iter_widgets(self) -> Iterator[WidgetContribution]:
        """Iterate over discovered WidgetContributions."""
        for mf in self.iter_manifests(disabled=False):
            yield from mf.contributions.widgets or ()

    def iter_sample_data(
        self,
    ) -> Iterator[Tuple[PluginName, List[SampleDataContribution]]]:
        """Iterates over (plugin_name, [sample_contribs])."""
        for mf in self.iter_manifests(disabled=False):
            if mf.contributions.sample_data:
                yield mf.name, mf.contributions.sample_data

    def get_writer(
        self, path: str, layer_types: Sequence[str], plugin_name: Optional[str] = None
    ) -> Tuple[Optional[WriterContribution], str]:
        """Get Writer contribution appropriate for `path`, and `layer_types`.

        When `path` has a file extension, find a compatible writer that has
        that same extension. When there is no extension and only a single layer,
        find a compatible writer and append the extension.
        Otherwise, find a compatible no-extension writer and write to that.
        No-extension writers typically write to a folder.

        Parameters
        ----------
        path : str
            Path to write
        layer_types : Sequence[str]
            Sequence of layer type strings (e.g. ['image', 'labels'])
        plugin_name : Optional[str], optional
            Name of plugin to use. If provided, only writers from `plugin_name` will be
            considered, otherwise all plugins are considered. by default `None`.

        Returns
        -------
        Tuple[Optional[WriterContribution], str]
            WriterContribution and path that will be written.
        """
        ext = Path(path).suffix.lower() if path else ""

        for writer in self.iter_compatible_writers(layer_types):
            if not plugin_name or writer.command.startswith(plugin_name):
                if (
                    ext
                    and ext in writer.filename_extensions
                    or not ext
                    and len(layer_types) != 1
                    and not writer.filename_extensions
                ):
                    return writer, path
                elif not ext and len(layer_types) == 1:  # No extension, single layer.
                    ext = next(iter(writer.filename_extensions), "")
                    return writer, path + ext
                # When the list of extensions for the writer doesn't match the
                # extension in the filename, keep searching.

        # Nothing got found
        return None, path


class PluginContext:
    """An object that can contain information for a plugin over its lifetime."""

    # stores all created contexts (currently cleared by `PluginManager.deactivate`)

    def __init__(
        self, plugin_key: PluginName, reg: Optional[CommandRegistry] = None
    ) -> None:
        self._activated = False
        self.plugin_key = plugin_key
        self._command_registry = reg or PluginManager.instance().commands
        self._imports: Set[str] = set()  # modules that were imported by this plugin
        # functions to call when deactivating
        self._disposables: Set[DisposeFunction] = set()

    def _dispose(self):
        while self._disposables:
            self._disposables.pop()()

    def register_command(self, id: str, command: Optional[Callable] = None):
        """Associate a callable with a command id."""

        def _inner(command):
            self._disposables.add(self._command_registry.register(id, command))
            return command

        return _inner if command is None else _inner(command)

    def register_disposable(self, func: DisposeFunction):
        """Register `func` to be executed when this plugin is deactivated."""
        self._disposables.add(func)


def _call_python_name(python_name: PythonName, args=()) -> Any:
    """convenience to call `python_name` function. eg `module.submodule:funcname`."""
    from .manifest.utils import import_python_name

    if not python_name:  # pragma: no cover
        return None

    func = import_python_name(python_name)
    if callable(func):
        return func(*args)


def _expand_dotted_set(inclusion_set: InclusionSet) -> InclusionSet:
    """Expand a set of strings with dots to a dict of dicts.

    Examples
    --------
    >>> _expand_dotted_set({'a.b', 'c', 'a.d'})
    {'a': {'b': True, 'd': True}, 'c': True}

    >>> _expand_dotted_set({'a.b', 'a.d.e', 'a'})
    {'a'}

    >>> _expand_dotted_set({'a.b', 'a.d', 'x.y.z'})
    {'x': {'y': {'z': True}}, 'a': {'d': True, 'b': True}}
    """
    if not isinstance(inclusion_set, set) or all(
        "." not in str(s) for s in inclusion_set
    ):
        return inclusion_set

    result: Dict[IntStr, Any] = {}
    # sort the strings based on the number of dots,
    # so that higher level keys take precedence
    # e.g. {'a.b', 'a.d.e', 'a'} -> {'a'}
    for key in sorted(inclusion_set, key=lambda i: i.count("."), reverse=True):
        if isinstance(key, str):
            parts = key.split(".")
            if len(parts) == 1:
                result[key] = True
            else:
                cur = result
                for part in parts[:-1]:
                    # integer keys are used in pydantic for lists
                    # they must remain integers
                    _p: IntStr = int(part) if part.isdigit() else part
                    cur = cur.setdefault(_p, {})
                cur[parts[-1]] = True

    return result



import os
import subprocess
import sys
import zipfile
from pathlib import Path

import pytest

from npe2 import PluginManifest

ROOT = Path(__file__).parent.parent

TEMPLATE = Path("my_module") / "_napari.yaml"
PYPROJECT = """
[build-system]
requires = ["setuptools", "wheel", "npe2 @ file://{}"]
build-backend = "setuptools.build_meta"

[tool.npe2]
template="{}"
""".format(
    ROOT, TEMPLATE
)


@pytest.mark.skipif(not os.getenv("CI"), reason="slow, only run on CI")
@pytest.mark.parametrize("dist_type", ["sdist", "wheel"])
def test_compile(compiled_plugin_dir: Path, tmp_path: Path, dist_type: str) -> None:
    """
    Test that the plugin manager can be compiled.
    """
    pyproject = compiled_plugin_dir / "pyproject.toml"
    pyproject.write_text(PYPROJECT)

    template = compiled_plugin_dir / TEMPLATE
    template.write_text("name: my_compiled_plugin\ndisplay_name: My Compiled Plugin\n")
    os.chdir(compiled_plugin_dir)
    subprocess.check_call([sys.executable, "-m", "build", f"--{dist_type}"])
    dist_dir = compiled_plugin_dir / "dist"
    assert dist_dir.is_dir()
    if dist_type == "sdist":
        # for sdist, test pip install into a temporary directory
        # and make sure the compiled manifest is there
        dist = next(dist_dir.glob("*.tar.gz"))
        site = tmp_path / "site"
        subprocess.check_call(
            [sys.executable, "-m", "pip", "install", str(dist), "--target", str(site)]
        )
        mf_file = site / "my_module" / "napari.yaml"
    else:
        # for wheel, make sure that the manifest is included in the wheel
        dist = next(dist_dir.glob("*.whl"))
        with zipfile.ZipFile(dist) as zip:
            zip.extractall(dist_dir)
        mf_file = dist_dir / "my_module" / "napari.yaml"

    assert mf_file.exists()
    mf = PluginManifest.from_file(mf_file)
    assert mf.display_name == "My Compiled Plugin"
    assert len(mf.contributions.readers) == 1
    assert len(mf.contributions.writers) == 2



import logging
import warnings
from typing import Optional, Union
from unittest.mock import patch

import pytest

from npe2 import DynamicPlugin, PluginManager, PluginManifest

logger = logging.getLogger(__name__)


class TestPluginManager(PluginManager):
    """A PluginManager subclass suitable for use in testing."""

    def discover(self, *_, **__) -> int:
        """Discovery is blocked in the TestPluginManager."""
        logger.warning(
            "NOTE: TestPluginManager refusing to discover plugins. "
            "You may add plugins to this test plugin manager using `tmp_plugin()`."
        )
        return 0

    def tmp_plugin(
        self,
        manifest: Optional[Union[PluginManifest, str]] = None,
        package: Optional[str] = None,
        name: Optional[str] = None,
    ) -> DynamicPlugin:
        """Create a DynamicPlugin instance using this plugin manager.

        If providing arguments, provide only one of 'manifest', 'package', 'name'.

        Parameters
        ----------
        manifest : Union[PluginManifest, str]
            A manifest to use for this plugin. If a string, it is assumed to be the
            path to a manifest file (which must exist), otherwise must be a
            PluginManifest instance.
        package : str
            Name of an installed plugin/package.
        name : str
            If neither `manifest` or `package` is provided, a new DynamicPlugin is
            created with this name, by default "tmp_plugin"

        Returns
        -------
        DynamicPlugin
            be sure to enter the DynamicPlugin context to register the plugin.

        Examples
        --------
        >>> def test_something_with_only_my_plugin_registered(npe2pm):
        ...    with npe2pm.tmp_plugin(package='my-plugin') as plugin:
        ...        ...

        >>> def test_something_with_specific_manifest_file_registered(npe2pm):
        ...    mf_file = Path(__file__).parent / 'my_manifest.yaml'
        ...    with npe2pm.tmp_plugin(manifest=str(mf_file)) as plugin:
        ...        ...
        """
        if manifest is not None:
            if package or name:  # pragma: no cover
                warnings.warn(
                    "`manifest` overrides the `package` and `name` arguments. "
                    "Please provide only one.",
                    stacklevel=2,
                )
            if isinstance(manifest, PluginManifest):
                mf = manifest
            else:
                mf = PluginManifest.from_file(manifest)
        elif package:
            if name:  # pragma: no cover
                warnings.warn(
                    "`package` overrides the `name` argument. Please provide only one.",
                    stacklevel=2,
                )
            mf = PluginManifest.from_distribution(package)
        else:
            name = name or "tmp_plugin"
            i = 0
            while name in self._manifests:  # pragma: no cover
                # guarantee that name is unique
                name = f"{name}_{i}"
                i += 1
            mf = PluginManifest(name=name)
        return DynamicPlugin(mf.name, plugin_manager=self, manifest=mf)


@pytest.fixture
def npe2pm():
    """Return mocked Global plugin manager instance, unable to discover plugins.

    Examples
    --------
    >>> @pytest.fixture(autouse=True)
    ... def mock_npe2_pm(npe2pm):
    ...     # Auto-use this fixture to prevent plugin discovery.
    ...     return npe2pm
    """
    _pm = TestPluginManager()
    with patch("npe2.PluginManager.instance", return_value=_pm):
        yield _pm



from functools import partial

import numpy as np
from magicgui import magic_factory
from napari_plugin_engine import napari_hook_implementation


class MyWidget:
    ...


def some_function(x: int):
    ...


def gen_data():
    ...


@napari_hook_implementation
def napari_get_reader(path):
    ...


@napari_hook_implementation
def napari_write_image(path, data, meta):
    ...


@napari_hook_implementation
def napari_write_labels(path, data, meta):
    ...


@napari_hook_implementation
def napari_provide_sample_data():
    return {
        "random data": gen_data,
        "local data": partial(np.ones, (4, 4)),
        "random image": "https://picsum.photos/1024",
        "sample_key": {
            "display_name": "Some Random Data (512 x 512)",
            "data": gen_data,
        },
        "local_ones": {
            "display_name": "Some local ones",
            "data": partial(np.ones, (4, 4)),
        },
    }


@napari_hook_implementation
def napari_experimental_provide_theme():
    return {
        "super_dark": {
            "name": "super_dark",
            "background": "rgb(12, 12, 12)",
            "foreground": "rgb(65, 72, 81)",
            "primary": "rgb(90, 98, 108)",
            "secondary": "rgb(134, 142, 147)",
            "highlight": "rgb(106, 115, 128)",
            "text": "rgb(240, 241, 242)",
            "icon": "rgb(209, 210, 212)",
            "warning": "rgb(153, 18, 31)",
            "current": "rgb(0, 122, 204)",
            "syntax_style": "native",
            "console": "rgb(0, 0, 0)",
            "canvas": "black",
        },
        "pretty_light": {
            "background": "rgb(192, 223, 139)",
        },
    }


factory = magic_factory(some_function)


@napari_hook_implementation
def napari_experimental_provide_dock_widget():
    @magic_factory
    def local_widget(y: str):
        ...

    return [
        MyWidget,
        (factory, {"name": "My Other Widget"}),
        (local_widget, {"name": "Local Widget"}),
    ]


@napari_hook_implementation
def napari_experimental_provide_function():
    def local_function(x: int):
        ...

    return [some_function, local_function]



from typing import Any, Dict, List, Literal, Optional, Union

from pydantic import BaseModel, Field, conlist, root_validator, validator

from ._json_schema import (
    Draft07JsonSchema,
    JsonType,
    JsonTypeArray,
    ValidationError,
    _coerce_type_name,
)


class ConfigurationProperty(Draft07JsonSchema):
    """Configuration for a single property in the plugin settings.

    This is a super-set of the JSON Schema (draft 7) specification.
    https://json-schema.org/understanding-json-schema/reference/index.html
    """

    type: Union[JsonType, JsonTypeArray] = Field(
        None,
        description="The type of this variable. Either JSON Schema type names ('array',"
        " 'boolean', 'object', ...) or python type names ('list', 'bool', 'dict', ...) "
        "may be used, but they will be coerced to JSON Schema types. Numbers, strings, "
        "and booleans will be editable in the UI, other types (lists, dicts) *may* be "
        "editable in the UI depending on their content, but maby will only be editable "
        "as text in the napari settings file. For boolean entries, the description "
        "(or markdownDescription) will be used as the label for the checkbox.",
    )
    _coerce_type_name = validator("type", pre=True, allow_reuse=True)(_coerce_type_name)

    default: Any = Field(None, description="The default value for this property.")

    description: Optional[str] = Field(
        None,
        description="Your `description` appears after the title and before the input "
        "field, except for booleans, where the description is used as the label for "
        "the checkbox. See also `markdown_description`.",
    )
    description_format: Literal["markdown", "plain"] = Field(
        "markdown",
        description="By default (`markdown`) your `description`, will be parsed "
        "as CommonMark (with `markdown_it`) and rendered as rich text. To render as "
        "plain text, set this value to `plain`.",
    )

    enum: Optional[conlist(Any, min_items=1, unique_items=True)] = Field(  # type: ignore # noqa: E501
        None,
        description="A list of valid options for this field. If you provide this field,"
        "the settings UI will render a dropdown menu.",
    )
    enum_descriptions: List[str] = Field(
        default_factory=list,
        description="If you provide a list of items under the `enum` field, you may "
        "provide `enum_descriptions` to add descriptive text for each enum.",
    )
    enum_descriptions_format: Literal["markdown", "plain"] = Field(
        "markdown",
        description="By default (`markdown`) your `enum_description`s, will be parsed "
        "as CommonMark (with `markdown_it`) and rendered as rich text. To render as "
        "plain text, set this value to `plain`.",
    )

    deprecation_message: Optional[str] = Field(
        None,
        description="If you set deprecationMessage, the setting will get a warning "
        "underline with your specified message. It won't show up in the settings "
        "UI unless it is configured by the user.",
    )
    deprecation_message_format: Literal["markdown", "plain"] = Field(
        "markdown",
        description="By default (`markdown`) your `deprecation_message`, will be "
        "parsed as CommonMark (with `markdown_it`) and rendered as rich text. To "
        "render as plain text, set this value to `plain`.",
    )

    edit_presentation: Literal["singleline", "multiline"] = Field(
        "singleline",
        description="By default, string settings will be rendered with a single-line "
        "editor. To render with a multi-line editor, set this value to `multiline`.",
    )
    order: Optional[int] = Field(
        None,
        description="When specified, gives the order of this setting relative to other "
        "settings within the same category. Settings with an order property will be "
        "placed before settings without this property set; and settings without `order`"
        " will be placed in alphabetical order.",
    )

    pattern_error_message: Optional[str] = Field(
        None,
        description="When restricting string types to a given regular expression with "
        "the `pattern` field, this field may be used to provide a custom error when "
        "the pattern does not match.",
    )

    @root_validator(pre=True)
    def _validate_root(cls, values):
        values = super()._validate_root(values)

        # we don't allow $ref and/or $defs in the schema
        for ignored in {"$ref", "ref", "definition", "$def"}:
            if ignored in values:
                import warnings

                del values[ignored]
                warnings.warn(
                    f"ignoring {ignored} in configuration property. "
                    "Configuration schemas must be self-contained.",
                    stacklevel=2,
                )
        return values

    def validate_instance(self, instance: object) -> dict:
        """Validate an object (instance) against this schema."""
        try:
            return super().validate_instance(instance)
        except ValidationError as e:
            if e.validator == "pattern" and self.pattern_error_message:
                e.message = self.pattern_error_message
            raise e


class ConfigurationContribution(BaseModel):
    """A configuration contribution for a plugin.

    This enables plugins to provide a schema for their configurables.
    Configuration contributions are used to generate the settings UI.
    """

    title: str = Field(
        ...,
        description="The heading used for this configuration category. Words like "
        '"Plugin", "Configuration", and "Settings" are redundant and should not be'
        "used in your title.",
    )
    properties: Dict[str, ConfigurationProperty] = Field(
        ...,
        description="Configuration properties. In the settings UI, your configuration "
        "key will be used to namespace and construct a title. Though a plugin can "
        "contain multiple categories of settings, each plugin setting must still have "
        "its own unique key. Capital letters in your key are used to indicate word "
        "breaks. For example, if your key is 'gitMagic.blame.dateFormat', the "
        "generated title for the setting will look like 'Blame: Date Format'",
    )
    # order: int  # vscode uses this to sort multiple configurations
    # ... I think we can just use the order in which they are declared



from npe2 import PackageMetadata


def test_package_metadata_version():
    """Test that we intelligently pick the min required metadata version"""
    assert PackageMetadata(name="test", version="1.0").metadata_version == "1.0"
    pm2 = PackageMetadata(name="test", version="1.0", maintainer="bob")
    assert pm2.metadata_version == "1.2"
    pm3 = PackageMetadata(
        name="test",
        version="1.0",
        maintainer="bob",
        description_content_type="text/markdown",
    )
    assert pm3.metadata_version == "2.1"


def test_hashable():
    hash(PackageMetadata(name="test", version="1.0"))


def test_package_metadata_extra_field():
    pkg = {
        "name": "test",
        "version": "1.0",
        "maintainer": "bob",
        "extra_field_that_is_definitely_not_in_the_model": False,
    }

    try:
        p = PackageMetadata(**pkg)
    except Exception as e:
        raise AssertionError(
            "failed to parse PackageMetadata from a dict with an extra field"
        ) from e

    assert p.name == "test"
    assert p.version == "1.0"
    assert p.maintainer == "bob"
    assert not hasattr(p, "extra_field_that_is_definitely_not_in_the_model")



import ast
import inspect
import re
import sys
import warnings
from configparser import ConfigParser
from functools import lru_cache, partial
from importlib import import_module, metadata
from logging import getLogger
from pathlib import Path
from types import ModuleType
from typing import (
    Any,
    Callable,
    DefaultDict,
    Dict,
    Iterator,
    List,
    Optional,
    Tuple,
    Union,
    cast,
)

from npe2.manifest import PluginManifest
from npe2.manifest.contributions import (
    CommandContribution,
    ThemeColors,
    WidgetContribution,
)
from npe2.manifest.utils import (
    SHIM_NAME_PREFIX,
    import_python_name,
    merge_manifests,
    safe_key,
)
from npe2.types import WidgetCreator

from ._setuputils import PackageInfo, get_package_dir_info

logger = getLogger(__name__)
NPE1_EP = "napari.plugin"
NPE2_EP = "napari.manifest"
NPE1_IMPL_TAG = "napari_impl"  # same as HookImplementation.format_tag("napari")


class HookImplementation:
    def __init__(
        self,
        function: Callable,
        plugin: Optional[ModuleType] = None,
        plugin_name: Optional[str] = None,
        **kwargs,
    ):
        self.function = function
        self.plugin = plugin
        self.plugin_name = plugin_name
        self._specname = kwargs.get("specname")

    def __repr__(self) -> str:  # pragma: no cover
        return (
            f"<HookImplementation plugin={self.plugin_name!r} spec={self.specname!r}>"
        )

    @property
    def specname(self) -> str:
        return self._specname or self.function.__name__


def iter_hookimpls(
    module: ModuleType, plugin_name: Optional[str] = None
) -> Iterator[HookImplementation]:
    # yield all routines in module that have "{self.project_name}_impl" attr
    for name in dir(module):
        method = getattr(module, name)
        if hasattr(method, NPE1_IMPL_TAG) and inspect.isroutine(method):
            hookimpl_opts = getattr(method, NPE1_IMPL_TAG)
            if isinstance(hookimpl_opts, dict):
                yield HookImplementation(method, module, plugin_name, **hookimpl_opts)


@lru_cache
def plugin_packages() -> List[PackageInfo]:
    """List of all packages with napari entry points.

    This is useful to help resolve naming issues (due to the terrible confusion
    around *what* a npe1 plugin name actually was).
    """

    packages: List[PackageInfo] = []
    for dist in metadata.distributions():
        packages.extend(
            PackageInfo(package_name=dist.metadata["Name"], entry_points=[ep])
            for ep in dist.entry_points
            if ep.group == NPE1_EP
        )

    return packages


def manifest_from_npe1(
    plugin: Union[str, metadata.Distribution, None] = None,
    module: Optional[Any] = None,
    adapter=False,
) -> PluginManifest:
    """Return manifest object given npe1 plugin or package name.

    One of `plugin` or `module` must be provide.

    Parameters
    ----------
    plugin : Union[str, metadata.Distribution, None]
        Name of package/plugin to convert.  Or a `metadata.Distribution` object.
        If a string, this function should be prepared to accept both the name of the
        package, and the name of an npe1 `napari.plugin` entry_point. by default None
    module : Optional[Module]
        namespace object, to directly import (mostly for testing.), by default None
    adapter : bool
        If True, the resulting manifest will be used internally by NPE1Adapter, but
        is NOT necessarily suitable for export as npe2 manifest. This will handle
        cases of locally defined functions and partials that don't have global
        python_names that are not supported natively by npe2. by default False
    """
    if module is not None:
        modules: List[str] = [module]
        package_name = "dynamic"
        plugin_name = getattr(module, "__name__", "dynamic_plugin")
    elif isinstance(plugin, str):
        modules = []
        plugin_name = plugin
        for pp in plugin_packages():
            if plugin in (pp.ep_name, pp.package_name):
                modules.append(pp.ep_value)
                package_name = pp.package_name
        if not modules:
            _avail = [f"  {p.package_name} ({p.ep_name})" for p in plugin_packages()]
            avail = "\n".join(_avail)
            raise metadata.PackageNotFoundError(
                f"No package or entry point found with name {plugin!r}: "
                f"\nFound packages (entry_point):\n{avail}"
            )
    elif hasattr(plugin, "entry_points") and hasattr(plugin, "metadata"):
        plugin = cast(metadata.Distribution, plugin)
        # don't use isinstance(Distribution), setuptools monkeypatches sys.meta_path:
        # https://github.com/pypa/setuptools/issues/3169
        NPE1_ENTRY_POINT = "napari.plugin"
        plugin_name = package_name = plugin.metadata["Name"]
        modules = [
            ep.value for ep in plugin.entry_points if ep.group == NPE1_ENTRY_POINT
        ]
        assert modules, f"No npe1 entry points found in distribution {plugin_name!r}"
    else:
        raise ValueError("one of plugin or module must be provided")  # pragma: no cover

    manifests: List[PluginManifest] = []
    for mod_name in modules:
        logger.debug(
            "Discovering contributions for npe1 plugin %r: module %r",
            package_name,
            mod_name,
        )
        parser = HookImplParser(package_name, plugin_name or "", adapter=adapter)
        _mod = import_module(mod_name) if isinstance(mod_name, str) else mod_name
        parser.parse_module(_mod)
        manifests.append(parser.manifest())

    assert manifests, "No npe1 entry points found in distribution {name}"
    return merge_manifests(manifests)


class HookImplParser:
    def __init__(self, package: str, plugin_name: str, adapter: bool = False) -> None:
        """A visitor class to convert npe1 hookimpls to a npe2 manifest

        Parameters
        ----------
        package : str
            Name of package
        plugin_name : str
            Name of plugin (will almost always be name of package)
        adapter : bool, optional
            If True, the resulting manifest will be used internally by NPE1Adapter, but
            is NOT necessarily suitable for export as npe2 manifest. This will handle
            cases of locally defined functions and partials that don't have global
            python_names that are not supported natively by npe2. by default False

        Examples
        --------
        >>> parser = HookImplParser(package, plugin_name)
        >>> parser.parse_callers(plugin_manager._plugin2hookcallers[_module])
        >>> mf = PluginManifest(name=package, contributions=dict(parser.contributions))
        """
        self.package = package
        self.plugin_name = plugin_name
        self.contributions: DefaultDict[str, list] = DefaultDict(list)
        self.adapter = adapter

    def manifest(self) -> PluginManifest:
        return PluginManifest(name=self.package, contributions=dict(self.contributions))

    def parse_module(self, module: ModuleType):
        for impl in iter_hookimpls(module, plugin_name=self.plugin_name):
            if impl.plugin_name == self.plugin_name:
                # call the corresponding hookimpl parser
                try:
                    getattr(self, impl.specname)(impl)
                except Exception as e:  # pragma: no cover
                    warnings.warn(
                        f"Failed to convert {impl.specname} in {self.package!r}: {e}",
                        stacklevel=2,
                    )

    def napari_experimental_provide_theme(self, impl: HookImplementation):
        ThemeDict = Dict[str, Union[str, Tuple, List]]
        d: Dict[str, ThemeDict] = impl.function()
        for name, theme_dict in d.items():
            colors = ThemeColors(**theme_dict)
            clr = colors.background or colors.foreground
            luma = _luma(*clr.as_rgb_tuple()[:3]) if clr else 0
            self.contributions["themes"].append(
                {
                    "label": name,
                    "id": name.lower().replace(" ", "_"),
                    "type": "dark" if luma < 128 else "light",
                    "colors": colors,
                }
            )

    def napari_get_reader(self, impl: HookImplementation):
        patterns = _guess_fname_patterns(impl.function)

        self.contributions["readers"].append(
            {
                "command": self.add_command(impl),
                "accepts_directories": True,
                "filename_patterns": patterns,
            }
        )

    def napari_provide_sample_data(self, impl: HookImplementation):
        module = sys.modules[impl.function.__module__.split(".", 1)[0]]

        samples: Dict[str, Union[dict, str, Callable]] = impl.function()
        for idx, (key, sample) in enumerate(samples.items()):
            _sample: Union[str, Callable]
            if isinstance(sample, dict):
                display_name = sample.get("display_name")
                _sample = sample.get("data")  # type: ignore
            else:
                _sample = sample
                display_name = key

            _key = safe_key(key)
            s = {"key": _key, "display_name": display_name}
            if callable(_sample):
                # let these raise exceptions here immediately if they don't validate
                id = f"{self.package}.data.{_key}"
                py_name = _python_name(
                    _sample, impl.function, hook_idx=idx if self.adapter else None
                )
                cmd_contrib = CommandContribution(
                    id=id,
                    python_name=py_name,
                    title=f"{key} sample",
                )
                self.contributions["commands"].append(cmd_contrib)
                s["command"] = id
            else:
                assert module.__file__
                package_dir = module.__file__.rsplit("/", 1)[0]
                s["uri"] = str(_sample).replace(package_dir, r"${package}")

            self.contributions["sample_data"].append(s)

    def napari_experimental_provide_function(self, impl: HookImplementation):
        items: Union[Callable, List[Callable]] = impl.function()
        items = [items] if not isinstance(items, list) else items

        for idx, item in enumerate(items):
            try:
                cmd = f"{self.package}.{item.__name__}"
                py_name = _python_name(
                    item, impl.function, hook_idx=idx if self.adapter else None
                )
                docsum = item.__doc__.splitlines()[0] if item.__doc__ else None
                cmd_contrib = CommandContribution(
                    id=cmd, python_name=py_name, title=docsum or item.__name__
                )
                self.contributions["commands"].append(cmd_contrib)

                wdg_contrib = WidgetContribution(
                    command=cmd,
                    display_name=item.__name__.replace("_", " "),
                    autogenerate=True,
                )
                self.contributions["widgets"].append(wdg_contrib)

            except Exception as e:  # pragma: no cover
                msg = (
                    f"Error converting function [{idx}] "
                    f"from {impl.function.__module__!r}:\n{e}"
                )
                warnings.warn(msg, stacklevel=2)

    def napari_experimental_provide_dock_widget(self, impl: HookImplementation):
        WidgetCallable = Union[Callable, Tuple[Callable, dict]]
        items: Union[WidgetCallable, List[WidgetCallable]] = impl.function()
        if not isinstance(items, list):
            items = [items]  # pragma: no cover

        # "wdg_creator" will be the function given by the plugin that returns a widget
        # while `impl` is the hook implementation that returned all the `wdg_creators`
        for idx, item in enumerate(items):
            if isinstance(item, tuple):
                wdg_creator = item[0]
                kwargs = item[1] if len(item) > 1 else {}
            else:
                wdg_creator, kwargs = (item, {})
            if not callable(wdg_creator) and isinstance(
                kwargs, dict
            ):  # pragma: no cover
                warnings.warn(
                    f"Invalid widget spec: {wdg_creator}, {kwargs}", stacklevel=2
                )
                continue

            try:
                func_name = getattr(wdg_creator, "__name__", "")
                wdg_name = str(kwargs.get("name", "")) or _camel_to_spaces(func_name)
                self._create_widget_contrib(
                    wdg_creator, display_name=wdg_name, idx=idx, hook=impl.function
                )
            except Exception as e:  # pragma: no cover
                msg = (
                    f"Error converting dock widget [{idx}] "
                    f"from {impl.function.__module__!r}:\n{e}"
                )
                warnings.warn(msg, stacklevel=2)

    def _create_widget_contrib(
        self,
        wdg_creator: WidgetCreator,
        display_name: str,
        idx: int,
        hook: Callable,
    ):
        # we provide both the wdg_creator object itself, as well as the hook impl that
        # returned it... In the case that we can't get an absolute python name to the
        # wdg_creator itself (e.g. it's defined in a local scope), then the py_name
        # will use the hookimpl itself, and the index of the object returned.
        py_name = _python_name(
            wdg_creator, hook, hook_idx=idx if self.adapter else None
        )

        if not py_name:  # pragma: no cover
            raise ValueError(
                "No suitable python name to point to. "
                "Is this a locally defined function or partial?"
            )

        func_name = getattr(wdg_creator, "__name__", "")
        cmd = f"{self.package}.{func_name or display_name.lower().replace(' ', '_')}"

        # let these raise exceptions here immediately if they don't validate
        cmd_contrib = CommandContribution(
            id=cmd, python_name=py_name, title=f"Create {display_name}"
        )
        wdg_contrib = WidgetContribution(command=cmd, display_name=display_name)
        self.contributions["commands"].append(cmd_contrib)
        self.contributions["widgets"].append(wdg_contrib)

    def napari_get_writer(self, impl: HookImplementation):
        warnings.warn(
            f"Found a multi-layer writer in {self.package!r} - {impl.specname!r}, "
            "but it's not convertable. Please add the writer manually.",
            stacklevel=2,
        )
        return NotImplemented  # pragma: no cover

    def napari_write_image(self, impl: HookImplementation):
        self._parse_writer(impl, "image")

    def napari_write_labels(self, impl: HookImplementation):
        self._parse_writer(impl, "labels")

    def napari_write_points(self, impl: HookImplementation):
        self._parse_writer(impl, "points")

    def napari_write_shapes(self, impl: HookImplementation):
        self._parse_writer(impl, "shapes")

    def napari_write_vectors(self, impl: HookImplementation):
        self._parse_writer(impl, "vectors")

    def _parse_writer(self, impl: HookImplementation, layer: str):
        id = self.add_command(impl)
        self.contributions["writers"].append(
            {
                "command": id,
                "layer_types": [layer],
                "display_name": layer,
                "filename_extensions": [],
            }
        )

    def add_command(self, impl: HookImplementation, py_name: str = "") -> str:
        name = impl.specname.replace("napari_", "")
        id = f"{self.package}.{name}"
        title = " ".join(name.split("_")).title()
        if not py_name:
            py_name = _python_name(impl.function)
        c = CommandContribution(id=id, python_name=py_name, title=title)
        self.contributions["commands"].append(c)
        return id


def _is_magicgui_magic_factory(obj):
    return "magicgui" in sys.modules and isinstance(obj, partial)


def _python_name(
    obj: Any, hook: Optional[Callable] = None, hook_idx: Optional[int] = None
) -> str:
    """Get resolvable python name for `obj` returned from an npe1 `hook` implentation.

    Parameters
    ----------
    obj : Any
        a python obj
    hook : Callable, optional
        the npe1 hook implementation that returned `obj`, by default None.
        This is used both to search the module namespace for `obj`, and also
        in the shim python name if `obj` cannot be found.
    hook_idx : int, optional
        If `obj` cannot be found and `hook_idx` is not None, then a shim name.
        of the form "__npe1shim__.{_python_name(hook)}_{hook_idx}" will be returned.
        by default None.

    Returns
    -------
    str
       a string that can be imported with npe2.manifest.utils.import_python_name

    Raises
    ------
    AttributeError
        If a resolvable string cannot be found
    """
    obj_name: Optional[str] = None
    mod_name: Optional[str] = None
    # first, check the global namespace of the module where the hook was declared
    # if we find `obj` itself, we can just use it.
    if hasattr(hook, "__module__"):
        hook_mod = sys.modules.get(hook.__module__)
        if hook_mod:
            for local_name, _obj in vars(hook_mod).items():
                if _obj is obj:
                    obj_name = local_name
                    mod_name = hook_mod.__name__
                    break

    # trick if it's a magic_factory
    if _is_magicgui_magic_factory(obj):
        f = obj.keywords.get("function")
        if f:
            v = getattr(f, "__globals__", {}).get(getattr(f, "__name__", ""))
            if v is obj:  # pragma: no cover
                mod_name = f.__module__
                obj_name = f.__qualname__

    # if that didn't work get the qualname of the object
    # and, if it's not a locally defined qualname, get the name of the module
    # in which it is defined
    if not (mod_name and obj_name):
        obj_name = getattr(obj, "__qualname__", "")
        if obj_name and "<locals>" not in obj_name:
            mod = inspect.getmodule(obj) or inspect.getmodule(hook)
            if mod:
                mod_name = mod.__name__

    if not (mod_name and obj_name) and (hook and hook_idx is not None):
        # we weren't able to resolve an absolute name... if we are shimming, then we
        # can create a special py_name of the form `__npe1shim__.hookfunction_idx`
        return f"{SHIM_NAME_PREFIX}{_python_name(hook)}_{hook_idx}"

    if obj_name and "<locals>" in obj_name:
        raise ValueError("functions defined in local scopes are not yet supported.")
    if not mod_name:
        raise AttributeError(f"could not get resolvable python name for {obj}")
    pyname = f"{mod_name}:{obj_name}"
    if import_python_name(pyname) is not obj:  # pragma: no cover
        raise AttributeError(f"could not get resolvable python name for {obj}")
    return pyname


def _luma(r, g, b):
    # https://en.wikipedia.org/wiki/Luma_(video)
    return 0.2126 * r + 0.7152 * g + 0.0722 * b  # per ITU-R BT.709


_camel_to_spaces_pattern = re.compile(r"((?<=[a-z])[A-Z]|(?<!\A)[A-R,T-Z](?=[a-z]))")


def _camel_to_spaces(val):
    return _camel_to_spaces_pattern.sub(r" \1", val)


def get_top_module_path(package_name, top_module: Optional[str] = None) -> Path:
    dist = metadata.distribution(package_name)
    if not top_module:
        top_mods = (dist.read_text("top_level.txt") or "").strip().splitlines()
        if not top_mods:
            raise ValueError(  # pragma: no cover
                "Could not detect a top level module in distribution metadata "
                f"for {package_name}"
            )
        top_module = top_mods[0]

    path = Path(dist.locate_file(top_module))
    assert path.is_dir()
    return path


def convert_repository(
    path: Union[Path, str], mf_name: str = "napari.yaml", dry_run=False
) -> Tuple[PluginManifest, Path]:
    """Convert repository at `path` to new npe2 style."""
    path = Path(path)

    # get the info we need and create a manifest
    info = get_package_dir_info(path)
    if not (info.package_name and info._ep1):
        msg = f'Could not detect first gen napari plugin package at "{path}".'
        if info._ep2 is not None:
            msg += f" Found a {NPE2_EP} entry_point. Is this package already converted?"
        raise ValueError(msg)

    manifest = manifest_from_npe1(info.package_name)
    top_module = get_top_module_path(info.package_name, info.top_module)
    if not top_module.is_dir():
        raise ValueError(  # pragma: no cover
            f"Detection of top-level module failed. {top_module} is not a directory."
        )
    mf_path = top_module / mf_name

    if dry_run:
        return manifest, mf_path

    # update the entry_points in setup.cfg/setup.py
    if info.setup_cfg:
        _write_new_setup_cfg_ep(info, mf_name)
    # or tell them to do it themselves in setup.py
    else:
        # tell them to do it manually
        warnings.warn(
            "\nCannot auto-update setup.py, please edit setup.py as follows:\n"
            "  1. remove the `napari.plugin` entry_point\n"
            "  2. add the following entry_point:"
            f"""
       entry_points={{
           "{NPE2_EP}": [
               "{info.package_name} = {info.top_module}:{mf_name}",
           ],
       }},
       package_data={{"{info.top_module}": ["{mf_name}"]}},
""",
            stacklevel=2,
        )

    # write the yaml to top_module/napari.yaml
    mf_path.write_text(manifest.yaml())
    return manifest, mf_path


def _write_new_setup_cfg_ep(info: PackageInfo, mf_name: str):
    assert info.setup_cfg
    p = ConfigParser(comment_prefixes="/", allow_no_value=True)  # preserve comments
    p.read(info.setup_cfg)
    mf_path = f"{info.top_module}:{mf_name}"
    new_ep = f"\n{info.package_name} = {mf_path}"
    if "options.entry_points" not in p.sections():
        p.add_section("options.entry_points")  # pragma: no cover
    p.set("options.entry_points", NPE2_EP, new_ep)
    if "options.package_data" not in p.sections():
        p.add_section("options.package_data")
    p.set("options.package_data", info.top_module, mf_name)
    if "options" not in p.sections():
        p.add_section("options")
    p.set("options", "include_package_data", "True")
    p.remove_option("options.entry_points", NPE1_EP)
    with open(info.setup_cfg, "w") as fh:
        p.write(fh)


def _guess_fname_patterns(func):
    """Try to guess filename extension patterns from source code.  Fallback to "*"."""

    patterns = ["*"]
    # try to look at source code to guess file extensions
    _, *b = inspect.getsource(func).split("endswith(")
    if b:
        try:
            middle = b[0].split(")")[0]
            if middle.startswith("("):
                middle += ")"
            files = ast.literal_eval(middle)
            if isinstance(files, str):
                files = [files]
            if files:
                patterns = [f"*{f}" for f in files]
        except Exception:  # pragma: no cover
            # couldn't do it... just accept all filename patterns
            pass
    return patterns



import json
from contextlib import contextmanager
from pathlib import Path
from typing import Callable, Dict, Optional, Union

import pytomlpp as toml
import yaml
from pydantic import BaseModel, PrivateAttr


class ImportExportModel(BaseModel):
    """Model mixin/base class that provides read/write from toml/yaml/json.

    To force the inclusion of a given field in the exported toml/yaml use:

        class MyModel(ImportExportModel):
            some_field: str = Field(..., always_export=True)
    """

    _source_file: Optional[Path] = PrivateAttr(None)

    def toml(self, pyproject=False, **kwargs) -> str:
        """Generate serialized `toml` string for this model.

        Parameters
        ----------
        pyproject : bool, optional
            If `True`, output will be in pyproject format, with all data under
            `tool.napari`, by default `False`.
        **kwargs
            passed to `BaseModel.json()`
        """
        d = self._serialized_data(**kwargs)
        if pyproject:
            d = {"tool": {"napari": d}}
        return toml.dumps(d)

    def yaml(self, **kwargs) -> str:
        """Generate serialized `yaml` string for this model.

        Parameters
        ----------
        **kwargs
            passed to `BaseModel.json()`
        """
        return yaml.safe_dump(self._serialized_data(**kwargs), sort_keys=False)

    @classmethod
    def from_file(cls, path: Union[Path, str]):
        """Parse model from a metadata file.

        Parameters
        ----------
        path : Path or str
            Path to file.  Must have extension {'.json', '.yaml', '.yml', '.toml'}

        Returns
        -------
        object
            The parsed model.

        Raises
        ------
        FileNotFoundError
            If `path` does not exist.
        ValueError
            If the file extension is not in {'.json', '.yaml', '.yml', '.toml'}
        """
        path = Path(path).expanduser().absolute().resolve()
        if not path.exists():
            raise FileNotFoundError(f"File not found: {path}")

        loader: Callable
        if path.suffix.lower() == ".json":
            loader = json.load
        elif path.suffix.lower() == ".toml":
            loader = toml.load
        elif path.suffix.lower() in (".yaml", ".yml"):
            loader = yaml.safe_load
        else:
            raise ValueError(f"unrecognized file extension: {path}")  # pragma: no cover

        with open(path, encoding="utf-8") as f:
            data = loader(f) or {}

        if path.name == "pyproject.toml":
            data = data["tool"]["napari"]

        obj = cls(**data)
        obj._source_file = Path(path).expanduser().absolute().resolve()
        return obj

    def _serialized_data(self, **kwargs):
        """using json encoders for all outputs"""
        kwargs.setdefault("exclude_unset", True)
        with self._required_export_fields_set():
            return json.loads(self.json(**kwargs))

    @contextmanager
    def _required_export_fields_set(self):
        fields = self.__fields__.items()
        required = {k for k, v in fields if v.field_info.extra.get("always_export")}

        was_there: Dict[str, bool] = {}
        for f in required:
            was_there[f] = f in self.__fields_set__
            self.__fields_set__.add(f)
        try:
            yield
        finally:
            for f in required:
                if not was_there.get(f):
                    self.__fields_set__.discard(f)



import os
import urllib.request
from importlib.metadata import PackageNotFoundError
from unittest.mock import patch

import pytest

from npe2 import PluginManifest, fetch_manifest
from npe2._inspection._fetch import (
    _manifest_from_pypi_sdist,
    get_hub_plugin,
    get_hub_plugins,
    get_manifest_from_wheel,
    get_pypi_plugins,
    get_pypi_url,
)
from npe2._inspection._full_install import fetch_manifest_with_full_install
from npe2.manifest._npe1_adapter import NPE1Adapter


def test_fetch_npe2_manifest():
    mf = fetch_manifest("napari-omero")
    assert mf.name == "napari-omero"
    assert any(mf.contributions.dict().values())
    assert mf.npe1_shim is False


def test_fetch_npe1_manifest_with_writer():
    mf = fetch_manifest("example-plugin")
    assert mf.name == "example-plugin"
    assert mf.contributions.writers
    # Test will eventually fail when example-plugin is updated to npe2
    # This is here as a sentinel
    assert mf.npe1_shim is True


def test_fetch_npe1_manifest_with_sample_data():
    mf = fetch_manifest("napari-pyclesperanto-assistant")
    assert mf.name == "napari-pyclesperanto-assistant"
    assert mf.contributions.sample_data
    # Test will eventually fail when napari-pyclesperanto-assistant is updated to npe2
    # This is here as a sentinel
    assert mf.npe1_shim is True


def test_fetch_npe1_manifest_dock_widget_as_attribute():
    # This tests is just to add coverage of a specific branch of code in the
    # napari_experimental_provide_dock_widget parser, (where the return value
    # is a dotted attribute, rather than a direct name).  I only saw it in
    # brainreg-segment.
    mf = fetch_manifest("brainreg-segment")
    assert mf.name == "brainreg-segment"
    assert mf.contributions.widgets
    # Test will eventually fail when brainreg-segment is updated to npe2
    # This is here as a sentinel
    assert mf.npe1_shim is True


@pytest.mark.parametrize("version", [None, "0.1.0"])
@pytest.mark.parametrize("packagetype", ["sdist", "bdist_wheel", None])
def test_get_pypi_url(version, packagetype):
    assert "npe2" in get_pypi_url("npe2", version=version, packagetype=packagetype)


def test_from_pypi_wheel_bdist_missing():
    error = PackageNotFoundError("No bdist_wheel releases found")
    with patch("npe2._inspection._fetch.get_pypi_url", side_effect=error):
        with pytest.raises(PackageNotFoundError):
            fetch_manifest("my-package")


@pytest.mark.skipif(not os.getenv("CI"), reason="slow, only run on CI")
def testfetch_manifest_with_full_install():
    # TODO: slowest of the tests ... would be nice to provide a local mock
    mf = fetch_manifest_with_full_install("napari-ndtiffs", version="0.1.2")
    # use version 0.1.2 which is npe1
    assert isinstance(mf, NPE1Adapter)
    assert mf.name == "napari-ndtiffs"
    assert mf.contributions


@pytest.mark.skipif(not os.getenv("CI"), reason="slow, only run on CI")
def test_manifest_from_sdist():
    mf = _manifest_from_pypi_sdist("zarpaint")
    assert mf.name == "zarpaint"


def test_get_manifest_from_wheel(tmp_path):
    url = "https://files.pythonhosted.org/packages/f0/cc/7f6fbce81be3eb73266f398e49df92859ba247134eb086704dd70b43819a/affinder-0.2.3-py3-none-any.whl"  # noqa
    dest = tmp_path / "affinder-0.2.3-py3-none-any.whl"
    urllib.request.urlretrieve(url, dest)
    mf = get_manifest_from_wheel(dest)
    assert mf.name == "affinder"


def test_get_hub_plugins():
    plugins = get_hub_plugins()
    assert len(plugins) > 0


def test_get_hub_plugin():
    info = get_hub_plugin("napari-svg")
    assert info["name"] == "napari-svg"


def test_get_pypi_plugins():
    plugins = get_pypi_plugins()
    assert len(plugins) > 0


@pytest.mark.skipif(not os.getenv("CI"), reason="slow, only run on CI")
@pytest.mark.parametrize(
    "url",
    [
        "https://files.pythonhosted.org/packages/fb/01/e59bc1d6ac96f84ce9d7a46cc5422250e047958ead6c5693ed386cf94003/napari_dv-0.3.0.tar.gz",  # noqa
        "https://files.pythonhosted.org/packages/5d/ae/17779e12ce60d8329306963e1a8dec608465caee582440011ff0c1310715/example_plugin-0.0.7-py3-none-any.whl",  # noqa
        "git+https://github.com/DragaDoncila/example-plugin.git",
        # this one doesn't use setuptools_scm, can check direct zip without clone
        "https://github.com/jo-mueller/napari-stl-exporter/archive/refs/heads/main.zip",
    ],
)
def test_fetch_urls(url):
    assert isinstance(fetch_manifest(url), PluginManifest)
